@@ -1,6 +1,7 @@
 /*
 Minetest
 Copyright (C) 2010-2013 celeron55, Perttu Ahola <celeron55@gmail.com>
+Copyright (C) 2020 numzero, Lobachevskiy Vitaliy <numzer0@yandex.ru>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
@@ -17,100 +18,92 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
+#include <cmath>
 #include "sky.h"
-#include "ITexture.h"
-#include "IVideoDriver.h"
-#include "ISceneManager.h"
-#include "ICameraSceneNode.h"
-#include "S3DVertex.h"
+#include <ITexture.h>
+#include <IVideoDriver.h>
+#include <ISceneManager.h>
+#include <ICameraSceneNode.h>
+#include <S3DVertex.h>
 #include "client/tile.h"
 #include "noise.h" // easeCurve
 #include "profiler.h"
 #include "util/numeric.h"
-#include <cmath>
 #include "client/renderingengine.h"
 #include "settings.h"
 #include "camera.h" // CameraModes
-#include "config.h"
+
 using namespace irr::core;
 
-Sky::Sky(s32 id, ITextureSource *tsrc) :
-		scene::ISceneNode(RenderingEngine::get_scene_manager()->getRootSceneNode(),
-			RenderingEngine::get_scene_manager(), id)
+static video::SMaterial baseMaterial()
 {
+	video::SMaterial mat;
+	mat.Lighting = false;
+	mat.ZBuffer = video::ECFN_DISABLED;
+	mat.ZWriteEnable = video::EZW_OFF;
+	mat.AntiAliasing = 0;
+	mat.TextureLayers[0].TextureWrapU = video::ETC_CLAMP_TO_EDGE;
+	mat.TextureLayers[0].TextureWrapV = video::ETC_CLAMP_TO_EDGE;
+	mat.BackfaceCulling = false;
+	return mat;
+}
+
+static inline void disableTextureFiltering(video::SMaterial &mat)
+{
+	mat.forEachTexture([] (auto &tex) {
+		tex.MinFilter = video::ETMINF_NEAREST_MIPMAP_NEAREST;
+		tex.MagFilter = video::ETMAGF_NEAREST;
+		tex.AnisotropicFilter = 0;
+	});
+}
+
+Sky::Sky(s32 id, RenderingEngine *rendering_engine, ITextureSource *tsrc, IShaderSource *ssrc) :
+		scene::ISceneNode(rendering_engine->get_scene_manager()->getRootSceneNode(),
+			rendering_engine->get_scene_manager(), id)
+{
+	m_seed = (u64)myrand() << 32 | myrand();
+
 	setAutomaticCulling(scene::EAC_OFF);
 	m_box.MaxEdge.set(0, 0, 0);
 	m_box.MinEdge.set(0, 0, 0);
 
-	// Create material
+	m_enable_shaders = g_settings->getBool("enable_shaders");
 
-	video::SMaterial mat;
-	mat.Lighting = false;
-#if ENABLE_GLES
-	mat.ZBuffer = video::ECFN_DISABLED;
-#else
-	mat.ZBuffer = video::ECFN_NEVER;
-#endif
-	mat.ZWriteEnable = false;
-	mat.AntiAliasing = 0;
-	mat.TextureLayer[0].TextureWrapU = video::ETC_CLAMP_TO_EDGE;
-	mat.TextureLayer[0].TextureWrapV = video::ETC_CLAMP_TO_EDGE;
-	mat.BackfaceCulling = false;
+	m_sky_params = SkyboxDefaults::getSkyDefaults();
+	m_sun_params = SkyboxDefaults::getSunDefaults();
+	m_moon_params = SkyboxDefaults::getMoonDefaults();
+	m_star_params = SkyboxDefaults::getStarDefaults();
 
-	m_materials[0] = mat;
+	// Create materials
 
-	m_materials[1] = mat;
-	//m_materials[1].MaterialType = video::EMT_TRANSPARENT_VERTEX_ALPHA;
+	m_materials[0] = baseMaterial();
+	// FIXME: shouldn't this check m_enable_shaders?
+	m_materials[0].MaterialType = ssrc->getShaderInfo(ssrc->getShader("stars_shader", TILE_MATERIAL_ALPHA)).material;
+	m_materials[0].Lighting = true;
+	m_materials[0].ColorMaterial = video::ECM_NONE;
+
+	m_materials[1] = baseMaterial();
 	m_materials[1].MaterialType = video::EMT_TRANSPARENT_ALPHA_CHANNEL;
 
-	m_materials[2] = mat;
+	m_materials[2] = baseMaterial();
 	m_materials[2].setTexture(0, tsrc->getTextureForMesh("sunrisebg.png"));
 	m_materials[2].MaterialType = video::EMT_TRANSPARENT_ALPHA_CHANNEL;
-	//m_materials[2].MaterialType = video::EMT_TRANSPARENT_ADD_COLOR;
 
-	// Ensures that sun and moon textures and tonemaps are correct.
-	setSkyDefaults();
-	m_sun_texture = tsrc->isKnownSourceImage(m_sun_params.texture) ?
-		tsrc->getTextureForMesh(m_sun_params.texture) : NULL;
-	m_moon_texture = tsrc->isKnownSourceImage(m_moon_params.texture) ?
-		tsrc->getTextureForMesh(m_moon_params.texture) : NULL;
-	m_sun_tonemap = tsrc->isKnownSourceImage(m_sun_params.tonemap) ?
-		tsrc->getTexture(m_sun_params.tonemap) : NULL;
-	m_moon_tonemap = tsrc->isKnownSourceImage(m_moon_params.tonemap) ?
-		tsrc->getTexture(m_moon_params.tonemap) : NULL;
+	setSunTexture(m_sun_params.texture, m_sun_params.tonemap, tsrc);
 
-	if (m_sun_texture) {
-		m_materials[3] = mat;
-		m_materials[3].setTexture(0, m_sun_texture);
-		m_materials[3].MaterialType = video::EMT_TRANSPARENT_ALPHA_CHANNEL;
-		// Disables texture filtering
-		m_materials[3].setFlag(video::E_MATERIAL_FLAG::EMF_BILINEAR_FILTER, false);
-		m_materials[3].setFlag(video::E_MATERIAL_FLAG::EMF_TRILINEAR_FILTER, false);
-		m_materials[3].setFlag(video::E_MATERIAL_FLAG::EMF_ANISOTROPIC_FILTER, false);
-		// Use tonemaps if available
-		if (m_sun_tonemap)
-			m_materials[3].Lighting = true;
-	}
-	if (m_moon_texture) {
-		m_materials[4] = mat;
-		m_materials[4].setTexture(0, m_moon_texture);
-		m_materials[4].MaterialType = video::EMT_TRANSPARENT_ALPHA_CHANNEL;
-		// Disables texture filtering
-		m_materials[4].setFlag(video::E_MATERIAL_FLAG::EMF_BILINEAR_FILTER, false);
-		m_materials[4].setFlag(video::E_MATERIAL_FLAG::EMF_TRILINEAR_FILTER, false);
-		m_materials[4].setFlag(video::E_MATERIAL_FLAG::EMF_ANISOTROPIC_FILTER, false);
-		// Use tonemaps if available
-		if (m_moon_tonemap)
-			m_materials[4].Lighting = true;
-	}
+	setMoonTexture(m_moon_params.texture, m_moon_params.tonemap, tsrc);
 
 	for (int i = 5; i < 11; i++) {
-		m_materials[i] = mat;
+		m_materials[i] = baseMaterial();
 		m_materials[i].Lighting = true;
 		m_materials[i].MaterialType = video::EMT_SOLID;
 	}
+
 	m_directional_colored_fog = g_settings->getBool("directional_colored_fog");
-	setStarCount(1000, true);
+	m_sky_params.body_orbit_tilt = g_settings->getFloat("shadow_sky_body_orbit_tilt", -60., 60.);
+	m_sky_params.fog_start = rangelim(g_settings->getFloat("fog_start"), 0.0f, 0.99f);
+
+	setStarCount(1000);
 }
 
 void Sky::OnRegisterSceneNode()
@@ -129,7 +122,7 @@ void Sky::render()
 	if (!camera || !driver)
 		return;
 
-	ScopeProfiler sp(g_profiler, "Sky::render()", SPT_AVG);
+	ScopeProfiler sp(g_profiler, "Sky::render()", SPT_AVG, PRECISION_MICRO);
 
 	// Draw perspective skybox
 
@@ -144,7 +137,6 @@ void Sky::render()
 	driver->setTransform(video::ETS_WORLD, translate * scale);
 
 	if (m_sunlight_seen) {
-		float sunsize = 0.07;
 		video::SColorf suncolor_f(1, 1, 0, 1);
 		//suncolor_f.r = 1;
 		//suncolor_f.g = MYMAX(0.3, MYMIN(1.0, 0.7 + m_time_brightness * 0.5));
@@ -158,21 +150,10 @@ void Sky::render()
 		suncolor_f.g = MYMAX(0.3, MYMIN(1.0, 0.85 + m_time_brightness * 0.5));
 		suncolor_f.b = MYMAX(0.0, m_brightness);
 
-		float moonsize = 0.04;
 		video::SColorf mooncolor_f(0.50, 0.57, 0.65, 1);
 		video::SColorf mooncolor2_f(0.85, 0.875, 0.9, 1);
 
-		float nightlength = 0.415;
-		float wn = nightlength / 2;
-		float wicked_time_of_day = 0;
-		if (m_time_of_day > wn && m_time_of_day < 1.0 - wn)
-			wicked_time_of_day = (m_time_of_day - wn) / (1.0 - wn * 2) * 0.5 + 0.25;
-		else if (m_time_of_day < 0.5)
-			wicked_time_of_day = m_time_of_day / wn * 0.25;
-		else
-			wicked_time_of_day = 1.0 - ((1.0 - m_time_of_day) / wn * 0.25);
-		/*std::cerr<<"time_of_day="<<m_time_of_day<<" -> "
-				<<"wicked_time_of_day="<<wicked_time_of_day<<std::endl;*/
+		float wicked_time_of_day = getWickedTimeOfDay(m_time_of_day);
 
 		video::SColor suncolor = suncolor_f.toSColor();
 		video::SColor suncolor2 = suncolor2_f.toSColor();
@@ -202,7 +183,7 @@ void Sky::render()
 
 		const f32 t = 1.0f;
 		const f32 o = 0.0f;
-		static const u16 indices[4] = {0, 1, 2, 3};
+		static const u16 indices[6] = {0, 1, 2, 0, 2, 3};
 		video::S3DVertex vertices[4];
 
 		driver->setMaterial(m_materials[1]);
@@ -244,7 +225,7 @@ void Sky::render()
 						vertex.Pos.rotateXZBy(180);
 					}
 				}
-				driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+				driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 			}
 		}
 
@@ -252,11 +233,10 @@ void Sky::render()
 		if (m_visible) {
 			driver->setMaterial(m_materials[1]);
 			for (u32 j = 0; j < 4; j++) {
-				video::SColor c = cloudyfogcolor.getInterpolated(m_skycolor, 0.45);
-				vertices[0] = video::S3DVertex(-1, 0.08, -1, 0, 0, 1, c, t, t);
-				vertices[1] = video::S3DVertex( 1, 0.08, -1, 0, 0, 1, c, o, t);
-				vertices[2] = video::S3DVertex( 1, 0.12, -1, 0, 0, 1, c, o, o);
-				vertices[3] = video::S3DVertex(-1, 0.12, -1, 0, 0, 1, c, t, o);
+				vertices[0] = video::S3DVertex(-1, -0.02, -1, 0, 0, 1, m_bgcolor, t, t);
+				vertices[1] = video::S3DVertex( 1, -0.02, -1, 0, 0, 1, m_bgcolor, o, t);
+				vertices[2] = video::S3DVertex( 1, 0.45, -1, 0, 0, 1, m_skycolor, o, o);
+				vertices[3] = video::S3DVertex(-1, 0.45, -1, 0, 0, 1, m_skycolor, t, o);
 				for (video::S3DVertex &vertex : vertices) {
 					if (j == 0)
 						// Don't switch
@@ -271,31 +251,7 @@ void Sky::render()
 						// Switch from -Z (south) to +Z (north)
 						vertex.Pos.rotateXZBy(-180);
 				}
-				driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
-			}
-
-			// Draw far cloudy fog thing at and below all horizons
-			for (u32 j = 0; j < 4; j++) {
-				video::SColor c = cloudyfogcolor;
-				vertices[0] = video::S3DVertex(-1, -1.0, -1, 0, 0, 1, c, t, t);
-				vertices[1] = video::S3DVertex( 1, -1.0, -1, 0, 0, 1, c, o, t);
-				vertices[2] = video::S3DVertex( 1, 0.08, -1, 0, 0, 1, c, o, o);
-				vertices[3] = video::S3DVertex(-1, 0.08, -1, 0, 0, 1, c, t, o);
-				for (video::S3DVertex &vertex : vertices) {
-					if (j == 0)
-						// Don't switch
-						{}
-					else if (j == 1)
-						// Switch from -Z (south) to +X (east)
-						vertex.Pos.rotateXZBy(90);
-					else if (j == 2)
-						// Switch from -Z (south) to -X (west)
-						vertex.Pos.rotateXZBy(-90);
-					else
-						// Switch from -Z (south) to +Z (north)
-						vertex.Pos.rotateXZBy(-180);
-				}
-				driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+				driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 			}
 		}
 
@@ -326,16 +282,16 @@ void Sky::render()
 					// Switch from -Z (south) to -X (west)
 					vertex.Pos.rotateXZBy(-90);
 			}
-			driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+			driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 		}
 
 		// Draw sun
 		if (m_sun_params.visible)
-			draw_sun(driver, sunsize, suncolor, suncolor2, wicked_time_of_day);
+			draw_sun(driver, suncolor, suncolor2, wicked_time_of_day);
 
 		// Draw moon
 		if (m_moon_params.visible)
-			draw_moon(driver, moonsize, mooncolor, mooncolor2, wicked_time_of_day);
+			draw_moon(driver, mooncolor, mooncolor2, wicked_time_of_day);
 
 		// Draw far cloudy fog thing below all horizons in front of sun, moon
 		// and stars.
@@ -362,7 +318,7 @@ void Sky::render()
 						// Switch from -Z (south) to +Z (north)
 						vertex.Pos.rotateXZBy(-180);
 				}
-				driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+				driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 			}
 
 			// Draw bottom far cloudy fog thing in front of sun, moon and stars
@@ -371,7 +327,7 @@ void Sky::render()
 			vertices[1] = video::S3DVertex( 1, -1.0, -1, 0, 1, 0, c, o, t);
 			vertices[2] = video::S3DVertex( 1, -1.0, 1, 0, 1, 0, c, o, o);
 			vertices[3] = video::S3DVertex(-1, -1.0, 1, 0, 1, 0, c, t, o);
-			driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+			driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 		}
 	}
 }
@@ -401,20 +357,6 @@ void Sky::update(float time_of_day, float time_brightness,
 
 	bool is_dawn = (time_brightness >= 0.20 && time_brightness < 0.35);
 
-	/*
-	Development colours
-
-	video::SColorf bgcolor_bright_normal_f(170. / 255, 200. / 255, 230. / 255, 1.0);
-	video::SColorf bgcolor_bright_dawn_f(0.666, 200. / 255 * 0.7, 230. / 255 * 0.5, 1.0);
-	video::SColorf bgcolor_bright_dawn_f(0.666, 0.549, 0.220, 1.0);
-	video::SColorf bgcolor_bright_dawn_f(0.666 * 1.2, 0.549 * 1.0, 0.220 * 1.0, 1.0);
-	video::SColorf bgcolor_bright_dawn_f(0.666 * 1.2, 0.549 * 1.0, 0.220 * 1.2, 1.0);
-
-	video::SColorf cloudcolor_bright_dawn_f(1.0, 0.591, 0.4);
-	video::SColorf cloudcolor_bright_dawn_f(1.0, 0.65, 0.44);
-	video::SColorf cloudcolor_bright_dawn_f(1.0, 0.7, 0.5);
-	*/
-
 	video::SColorf bgcolor_bright_normal_f = m_sky_params.sky_color.day_horizon;
 	video::SColorf bgcolor_bright_indoor_f = m_sky_params.sky_color.indoors;
 	video::SColorf bgcolor_bright_dawn_f = m_sky_params.sky_color.dawn_horizon;
@@ -520,7 +462,7 @@ void Sky::update(float time_of_day, float time_brightness,
 
 			video::SColorf pointcolor_sun_f(1, 1, 1, 1);
 			// Use tonemap only if default sun/moon tinting is used
-			// which keeps previous behaviour.
+			// which keeps previous behavior.
 			if (m_sun_tonemap && m_default_tint) {
 				pointcolor_sun_f.r = pointcolor_light *
 					(float)m_materials[3].EmissiveColor.getRed() / 255;
@@ -529,7 +471,7 @@ void Sky::update(float time_of_day, float time_brightness,
 				pointcolor_sun_f.g = pointcolor_light *
 					(float)m_materials[3].EmissiveColor.getGreen() / 255;
 			} else if (!m_default_tint) {
-				pointcolor_sun_f = m_sky_params.sun_tint;
+				pointcolor_sun_f = m_sky_params.fog_sun_tint;
 			} else {
 				pointcolor_sun_f.r = pointcolor_light * 1;
 				pointcolor_sun_f.b = pointcolor_light *
@@ -548,9 +490,9 @@ void Sky::update(float time_of_day, float time_brightness,
 				);
 			} else {
 				pointcolor_moon_f = video::SColorf(
-					(m_sky_params.moon_tint.getRed() / 255) * pointcolor_light,
-					(m_sky_params.moon_tint.getGreen() / 255) * pointcolor_light,
-					(m_sky_params.moon_tint.getBlue() / 255) * pointcolor_light,
+					(m_sky_params.fog_moon_tint.getRed() / 255.0f) * pointcolor_light,
+					(m_sky_params.fog_moon_tint.getGreen() / 255.0f) * pointcolor_light,
+					(m_sky_params.fog_moon_tint.getBlue() / 255.0f) * pointcolor_light,
 					1
 				);
 			}
@@ -605,17 +547,38 @@ void Sky::update(float time_of_day, float time_brightness,
 	}
 }
 
-void Sky::draw_sun(video::IVideoDriver *driver, float sunsize, const video::SColor &suncolor,
+static v3f getSkyBodyPosition(float horizon_position, float day_position, float orbit_tilt)
+{
+	v3f result = v3f(0, 0, -1);
+	result.rotateXZBy(horizon_position);
+	result.rotateXYBy(day_position);
+	result.rotateYZBy(orbit_tilt);
+	return result;
+}
+
+v3f Sky::getSunDirection()
+{
+	return getSkyBodyPosition(90, getWickedTimeOfDay(m_time_of_day) * 360 - 90, m_sky_params.body_orbit_tilt);
+}
+
+v3f Sky::getMoonDirection()
+{
+	return getSkyBodyPosition(270, getWickedTimeOfDay(m_time_of_day) * 360 - 90, m_sky_params.body_orbit_tilt);
+}
+
+void Sky::draw_sun(video::IVideoDriver *driver, const video::SColor &suncolor,
 	const video::SColor &suncolor2, float wicked_time_of_day)
 	/* Draw sun in the sky.
 	 * driver: Video driver object used to draw
-	 * sunsize: the default size of the sun
 	 * suncolor: main sun color
 	 * suncolor2: second sun color
 	 * wicked_time_of_day: current time of day, to know where should be the sun in the sky
 	 */
 {
-	static const u16 indices[4] = {0, 1, 2, 3};
+	// A magic number that contributes to the ratio 1.57 sun/moon size difference.
+	constexpr float sunsize = 0.07;
+
+	static const u16 indices[] = {0, 1, 2, 0, 2, 3};
 	std::array<video::S3DVertex, 4> vertices;
 	if (!m_sun_texture) {
 		driver->setMaterial(m_materials[1]);
@@ -633,10 +596,12 @@ void Sky::draw_sun(video::IVideoDriver *driver, float sunsize, const video::SCol
 		for (int i = 0; i < 4; i++) {
 			draw_sky_body(vertices, -sunsizes[i], sunsizes[i], colors[i]);
 			place_sky_body(vertices, 90, wicked_time_of_day * 360 - 90);
-			driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+			driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 		}
 	} else {
 		driver->setMaterial(m_materials[3]);
+		// Another magic number that contributes to the ratio 1.57 sun/moon size
+		// difference.
 		float d = (sunsize * 1.7) * m_sun_params.scale;
 		video::SColor c;
 		if (m_sun_tonemap)
@@ -645,24 +610,26 @@ void Sky::draw_sun(video::IVideoDriver *driver, float sunsize, const video::SCol
 			c = video::SColor(255, 255, 255, 255);
 		draw_sky_body(vertices, -d, d, c);
 		place_sky_body(vertices, 90, wicked_time_of_day * 360 - 90);
-		driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+		driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 	}
 }
 
 
-void Sky::draw_moon(video::IVideoDriver *driver, float moonsize, const video::SColor &mooncolor,
+void Sky::draw_moon(video::IVideoDriver *driver, const video::SColor &mooncolor,
 	const video::SColor &mooncolor2, float wicked_time_of_day)
 /*
 	* Draw moon in the sky.
 	* driver: Video driver object used to draw
-	* moonsize: the default size of the moon
 	* mooncolor: main moon color
 	* mooncolor2: second moon color
 	* wicked_time_of_day: current time of day, to know where should be the moon in
 	* the sky
 	*/
 {
-	static const u16 indices[4] = {0, 1, 2, 3};
+	// A magic number that contributes to the ratio 1.57 sun/moon size difference.
+	constexpr float moonsize = 0.04;
+
+	static const u16 indices[] = {0, 1, 2, 0, 2, 3};
 	std::array<video::S3DVertex, 4> vertices;
 	if (!m_moon_texture) {
 		driver->setMaterial(m_materials[1]);
@@ -686,10 +653,12 @@ void Sky::draw_moon(video::IVideoDriver *driver, float moonsize, const video::SC
 		for (int i = 0; i < 4; i++) {
 			draw_sky_body(vertices, moonsizes_1[i], moonsizes_2[i], colors[i]);
 			place_sky_body(vertices, -90, wicked_time_of_day * 360 - 90);
-			driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+			driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 		}
 	} else {
 		driver->setMaterial(m_materials[4]);
+		// Another magic number that contributes to the ratio 1.57 sun/moon size
+		// difference.
 		float d = (moonsize * 1.9) * m_moon_params.scale;
 		video::SColor c;
 		if (m_moon_tonemap)
@@ -698,13 +667,12 @@ void Sky::draw_moon(video::IVideoDriver *driver, float moonsize, const video::SC
 			c = video::SColor(255, 255, 255, 255);
 		draw_sky_body(vertices, -d, d, c);
 		place_sky_body(vertices, -90, wicked_time_of_day * 360 - 90);
-		driver->drawIndexedTriangleFan(&vertices[0], 4, indices, 2);
+		driver->drawIndexedTriangleList(&vertices[0], 4, indices, 2);
 	}
 }
 
 void Sky::draw_stars(video::IVideoDriver * driver, float wicked_time_of_day)
 {
-	driver->setMaterial(m_materials[1]);
 	// Tune values so that stars first appear just after the sun
 	// disappears over the horizon, and disappear just before the sun
 	// appears over the horizon.
@@ -712,87 +680,22 @@ void Sky::draw_stars(video::IVideoDriver * driver, float wicked_time_of_day)
 	// to time 4000.
 
 	float tod = wicked_time_of_day < 0.5f ? wicked_time_of_day : (1.0f - wicked_time_of_day);
-	float starbrightness = clamp((0.25f - fabsf(tod)) * 20.0f, 0.0f, 1.0f);
+	float day_opacity = clamp(m_star_params.day_opacity, 0.0f, 1.0f);
+	float starbrightness = (0.25f - std::abs(tod)) * 20.0f;
+	float alpha = clamp(starbrightness, day_opacity, 1.0f);
 
-	float f = starbrightness;
-	float d = (0.006 / 2) * m_star_params.scale;
-
-	video::SColor starcolor = m_star_params.starcolor;
-	starcolor.setAlpha(f * m_star_params.starcolor.getAlpha());
-
-	// Stars are only drawn when not fully transparent
-	if (m_star_params.starcolor.getAlpha() < 1)
+	video::SColorf color(m_star_params.starcolor);
+	color.a *= alpha;
+	if (color.a <= 0.0f) // Stars are only drawn when not fully transparent
 		return;
-#if ENABLE_GLES
-	u16 *indices = new u16[m_star_params.count * 3];
-	video::S3DVertex *vertices =
-			new video::S3DVertex[m_star_params.count * 3];
-	for (u32 i = 0; i < m_star_params.count; i++) {
-		indices[i * 3 + 0] = i * 3 + 0;
-		indices[i * 3 + 1] = i * 3 + 1;
-		indices[i * 3 + 2] = i * 3 + 2;
-		v3f r = m_stars[i];
-		core::CMatrix4<f32> a;
-		a.buildRotateFromTo(v3f(0, 1, 0), r);
-		v3f p = v3f(-d, 1, -d);
-		v3f p1 = v3f(d, 1, 0);
-		v3f p2 = v3f(-d, 1, d);
-		a.rotateVect(p);
-		a.rotateVect(p1);
-		a.rotateVect(p2);
-		p.rotateXYBy(wicked_time_of_day * 360 - 90);
-		p1.rotateXYBy(wicked_time_of_day * 360 - 90);
-		p2.rotateXYBy(wicked_time_of_day * 360 - 90);
-		vertices[i * 3 + 0].Pos = p;
-		vertices[i * 3 + 0].Color = starcolor;
-		vertices[i * 3 + 1].Pos = p1;
-		vertices[i * 3 + 1].Color = starcolor;
-		vertices[i * 3 + 2].Pos = p2;
-		vertices[i * 3 + 2].Color = starcolor;
-	}
-	driver->drawIndexedTriangleList(vertices, m_star_params.count * 3,
-			indices, m_star_params.count);
-	delete[] indices;
-	delete[] vertices;
-#else
-	u16 *indices = new u16[m_star_params.count * 4];
-	video::S3DVertex *vertices =
-			new video::S3DVertex[m_star_params.count * 4];
-	for (u32 i = 0; i < m_star_params.count; i++) {
-		indices[i * 4 + 0] = i * 4 + 0;
-		indices[i * 4 + 1] = i * 4 + 1;
-		indices[i * 4 + 2] = i * 4 + 2;
-		indices[i * 4 + 3] = i * 4 + 3;
-		v3f r = m_stars[i];
-		core::CMatrix4<f32> a;
-		a.buildRotateFromTo(v3f(0, 1, 0), r);
-		v3f p = v3f(-d, 1, -d);
-		v3f p1 = v3f(d, 1, -d);
-		v3f p2 = v3f(d, 1, d);
-		v3f p3 = v3f(-d, 1, d);
-		a.rotateVect(p);
-		a.rotateVect(p1);
-		a.rotateVect(p2);
-		a.rotateVect(p3);
-		p.rotateXYBy(wicked_time_of_day * 360 - 90);
-		p1.rotateXYBy(wicked_time_of_day * 360 - 90);
-		p2.rotateXYBy(wicked_time_of_day * 360 - 90);
-		p3.rotateXYBy(wicked_time_of_day * 360 - 90);
-		vertices[i * 4 + 0].Pos = p;
-		vertices[i * 4 + 0].Color = starcolor;
-		vertices[i * 4 + 1].Pos = p1;
-		vertices[i * 4 + 1].Color = starcolor;
-		vertices[i * 4 + 2].Pos = p2;
-		vertices[i * 4 + 2].Color = starcolor;
-		vertices[i * 4 + 3].Pos = p3;
-		vertices[i * 4 + 3].Color = starcolor;
-	}
-	driver->drawVertexPrimitiveList(vertices, m_star_params.count * 4,
-			indices, m_star_params.count, video::EVT_STANDARD,
-			scene::EPT_QUADS, video::EIT_16BIT);
-	delete[] indices;
-	delete[] vertices;
-#endif
+	m_materials[0].EmissiveColor = color.toSColor();
+
+	auto sky_rotation = core::matrix4().setRotationAxisRadians(2.0f * M_PI * (wicked_time_of_day - 0.25f), v3f(0.0f, 0.0f, 1.0f));
+	auto world_matrix = driver->getTransform(video::ETS_WORLD);
+	driver->setTransform(video::ETS_WORLD, world_matrix * sky_rotation);
+	driver->setMaterial(m_materials[0]);
+	driver->drawMeshBuffer(m_stars.get());
+	driver->setTransform(video::ETS_WORLD, world_matrix);
 }
 
 void Sky::draw_sky_body(std::array<video::S3DVertex, 4> &vertices, float pos_1, float pos_2, const video::SColor &c)
@@ -821,50 +724,49 @@ void Sky::place_sky_body(
 	* day_position: turn the body around the Z axis, to place it depending of the time of the day
 	*/
 {
+	v3f centrum = getSkyBodyPosition(horizon_position, day_position, m_sky_params.body_orbit_tilt);
+	v3f untilted_centrum = getSkyBodyPosition(horizon_position, day_position, 0.f);
 	for (video::S3DVertex &vertex : vertices) {
 		// Body is directed to -Z (south) by default
 		vertex.Pos.rotateXZBy(horizon_position);
 		vertex.Pos.rotateXYBy(day_position);
+		vertex.Pos += centrum - untilted_centrum;
 	}
 }
 
-void Sky::setSunTexture(std::string sun_texture,
-		std::string sun_tonemap, ITextureSource *tsrc)
+void Sky::setSunTexture(const std::string &sun_texture,
+		const std::string &sun_tonemap, ITextureSource *tsrc)
 {
 	// Ignore matching textures (with modifiers) entirely,
 	// but lets at least update the tonemap before hand.
 	m_sun_params.tonemap = sun_tonemap;
-	m_sun_tonemap = tsrc->isKnownSourceImage(m_sun_params.tonemap) ?
-		tsrc->getTexture(m_sun_params.tonemap) : NULL;
+	m_sun_tonemap = tsrc->isKnownSourceImage(sun_tonemap) ?
+		tsrc->getTexture(sun_tonemap) : nullptr;
 	m_materials[3].Lighting = !!m_sun_tonemap;
 
-	if (m_sun_params.texture == sun_texture)
+	if (m_sun_params.texture == sun_texture && !m_first_update)
 		return;
 	m_sun_params.texture = sun_texture;
 
-	if (sun_texture != "") {
-		// We want to ensure the texture exists first.
-		m_sun_texture = tsrc->getTextureForMesh(m_sun_params.texture);
+	m_sun_texture = nullptr;
+	if (sun_texture == "sun.png") {
+		// Dumb compatibility fix: sun.png transparently falls back to no texture
+		m_sun_texture = tsrc->isKnownSourceImage(sun_texture) ?
+			tsrc->getTexture(sun_texture) : nullptr;
+	} else if (!sun_texture.empty()) {
+		m_sun_texture = tsrc->getTextureForMesh(sun_texture);
+	}
 
-		if (m_sun_texture) {
-			m_materials[3] = m_materials[0];
-			m_materials[3].setTexture(0, m_sun_texture);
-			m_materials[3].MaterialType = video::
-				EMT_TRANSPARENT_ALPHA_CHANNEL;
-			// Disables texture filtering
-			m_materials[3].setFlag(
-				video::E_MATERIAL_FLAG::EMF_BILINEAR_FILTER, false);
-			m_materials[3].setFlag(
-				video::E_MATERIAL_FLAG::EMF_TRILINEAR_FILTER, false);
-			m_materials[3].setFlag(
-				video::E_MATERIAL_FLAG::EMF_ANISOTROPIC_FILTER, false);
-		}
-	} else {
-		m_sun_texture = nullptr;
+	if (m_sun_texture) {
+		m_materials[3] = baseMaterial();
+		m_materials[3].setTexture(0, m_sun_texture);
+		m_materials[3].MaterialType = video::EMT_TRANSPARENT_ALPHA_CHANNEL;
+		disableTextureFiltering(m_materials[3]);
+		m_materials[3].Lighting = !!m_sun_tonemap;
 	}
 }
 
-void Sky::setSunriseTexture(std::string sunglow_texture,
+void Sky::setSunriseTexture(const std::string &sunglow_texture,
 		ITextureSource* tsrc)
 {
 	// Ignore matching textures (with modifiers) entirely.
@@ -876,73 +778,107 @@ void Sky::setSunriseTexture(std::string sunglow_texture,
 	);
 }
 
-void Sky::setMoonTexture(std::string moon_texture,
-		std::string moon_tonemap, ITextureSource *tsrc)
+void Sky::setMoonTexture(const std::string &moon_texture,
+	const std::string &moon_tonemap, ITextureSource *tsrc)
 {
 	// Ignore matching textures (with modifiers) entirely,
 	// but lets at least update the tonemap before hand.
 	m_moon_params.tonemap = moon_tonemap;
-	m_moon_tonemap = tsrc->isKnownSourceImage(m_moon_params.tonemap) ?
-		tsrc->getTexture(m_moon_params.tonemap) : NULL;
+	m_moon_tonemap = tsrc->isKnownSourceImage(moon_tonemap) ?
+		tsrc->getTexture(moon_tonemap) : nullptr;
 	m_materials[4].Lighting = !!m_moon_tonemap;
 
-	if (m_moon_params.texture == moon_texture)
+	if (m_moon_params.texture == moon_texture && !m_first_update)
 		return;
 	m_moon_params.texture = moon_texture;
 
-	if (moon_texture != "") {
-		// We want to ensure the texture exists first.
-		m_moon_texture = tsrc->getTextureForMesh(m_moon_params.texture);
+	m_moon_texture = nullptr;
+	if (moon_texture == "moon.png") {
+		// Dumb compatibility fix: moon.png transparently falls back to no texture
+		m_moon_texture = tsrc->isKnownSourceImage(moon_texture) ?
+			tsrc->getTexture(moon_texture) : nullptr;
+	} else if (!moon_texture.empty()) {
+		m_moon_texture = tsrc->getTextureForMesh(moon_texture);
+	}
 
-		if (m_moon_texture) {
-			m_materials[4] = m_materials[0];
-			m_materials[4].setTexture(0, m_moon_texture);
-			m_materials[4].MaterialType = video::
-				EMT_TRANSPARENT_ALPHA_CHANNEL;
-			// Disables texture filtering
-			m_materials[4].setFlag(
-				video::E_MATERIAL_FLAG::EMF_BILINEAR_FILTER, false);
-			m_materials[4].setFlag(
-				video::E_MATERIAL_FLAG::EMF_TRILINEAR_FILTER, false);
-			m_materials[4].setFlag(
-				video::E_MATERIAL_FLAG::EMF_ANISOTROPIC_FILTER, false);
-		}
-	} else {
-		m_moon_texture = nullptr;
+	if (m_moon_texture) {
+		m_materials[4] = baseMaterial();
+		m_materials[4].setTexture(0, m_moon_texture);
+		m_materials[4].MaterialType = video::EMT_TRANSPARENT_ALPHA_CHANNEL;
+		disableTextureFiltering(m_materials[4]);
+		m_materials[4].Lighting = !!m_moon_tonemap;
 	}
 }
 
-void Sky::setStarCount(u16 star_count, bool force_update)
+void Sky::setStarCount(u16 star_count)
 {
 	// Allow force updating star count at game init.
-	if (m_star_params.count != star_count || force_update) {
+	if (m_star_params.count != star_count || m_first_update) {
 		m_star_params.count = star_count;
-		m_stars.clear();
-		// Rebuild the stars surrounding the camera
-		for (u16 i = 0; i < star_count; i++) {
-			v3f star = v3f(
-				myrand_range(-10000, 10000),
-				myrand_range(-10000, 10000),
-				myrand_range(-10000, 10000)
-			);
-
-			star.normalize();
-			m_stars.emplace_back(star);
-		}
+		updateStars();
 	}
 }
 
-void Sky::setSkyColors(const SkyboxParams sky)
+void Sky::updateStars()
 {
-	m_sky_params.sky_color = sky.sky_color;
+	m_stars.reset(new scene::SMeshBuffer());
+	// Stupid IrrLicht doesn’t allow non-indexed rendering, and indexed quad
+	// rendering is slow due to lack of hardware support. So as indices are
+	// 16-bit and there are 4 vertices per star... the limit is 2^16/4 = 0x4000.
+	// That should be well enough actually.
+	if (m_star_params.count > 0x4000) {
+		warningstream << "Requested " << m_star_params.count << " stars but " << 0x4000 << " is the max\n";
+		m_star_params.count = 0x4000;
+	}
+	m_stars->Vertices.reallocate(4 * m_star_params.count);
+	m_stars->Indices.reallocate(6 * m_star_params.count);
+
+	video::SColor fallback_color = m_star_params.starcolor; // used on GLES 2 “without shaders”
+	PcgRandom rgen(m_seed);
+	float d = (0.006 / 2) * m_star_params.scale;
+	for (u16 i = 0; i < m_star_params.count; i++) {
+		v3f r = v3f(
+			rgen.range(-10000, 10000),
+			rgen.range(-10000, 10000),
+			rgen.range(-10000, 10000)
+		);
+		core::CMatrix4<f32> a;
+		a.buildRotateFromTo(v3f(0, 1, 0), r);
+		v3f p = v3f(-d, 1, -d);
+		v3f p1 = v3f(d, 1, -d);
+		v3f p2 = v3f(d, 1, d);
+		v3f p3 = v3f(-d, 1, d);
+		a.rotateVect(p);
+		a.rotateVect(p1);
+		a.rotateVect(p2);
+		a.rotateVect(p3);
+		m_stars->Vertices.push_back(video::S3DVertex(p, {}, fallback_color, {}));
+		m_stars->Vertices.push_back(video::S3DVertex(p1, {}, fallback_color, {}));
+		m_stars->Vertices.push_back(video::S3DVertex(p2, {}, fallback_color, {}));
+		m_stars->Vertices.push_back(video::S3DVertex(p3, {}, fallback_color, {}));
+	}
+	for (u16 i = 0; i < m_star_params.count; i++) {
+		m_stars->Indices.push_back(i * 4 + 0);
+		m_stars->Indices.push_back(i * 4 + 1);
+		m_stars->Indices.push_back(i * 4 + 2);
+		m_stars->Indices.push_back(i * 4 + 2);
+		m_stars->Indices.push_back(i * 4 + 3);
+		m_stars->Indices.push_back(i * 4 + 0);
+	}
+	m_stars->setHardwareMappingHint(scene::EHM_STATIC);
+}
+
+void Sky::setSkyColors(const SkyColor &sky_color)
+{
+	m_sky_params.sky_color = sky_color;
 }
 
 void Sky::setHorizonTint(video::SColor sun_tint, video::SColor moon_tint,
-		std::string use_sun_tint)
+	const std::string &use_sun_tint)
 {
 	// Change sun and moon tinting:
-	m_sky_params.sun_tint = sun_tint;
-	m_sky_params.moon_tint = moon_tint;
+	m_sky_params.fog_sun_tint = sun_tint;
+	m_sky_params.fog_moon_tint = moon_tint;
 	// Faster than comparing strings every rendering frame
 	if (use_sun_tint == "default")
 		m_default_tint = true;
@@ -952,7 +888,7 @@ void Sky::setHorizonTint(video::SColor sun_tint, video::SColor moon_tint,
 		m_default_tint = true;
 }
 
-void Sky::addTextureToSkybox(std::string texture, int material_id,
+void Sky::addTextureToSkybox(const std::string &texture, int material_id,
 		ITextureSource *tsrc)
 {
 	// Sanity check for more than six textures.
@@ -961,17 +897,21 @@ void Sky::addTextureToSkybox(std::string texture, int material_id,
 	// Keep a list of texture names handy.
 	m_sky_params.textures.emplace_back(texture);
 	video::ITexture *result = tsrc->getTextureForMesh(texture);
-	m_materials[material_id+5] = m_materials[0];
+	m_materials[material_id+5] = baseMaterial();
 	m_materials[material_id+5].setTexture(0, result);
 	m_materials[material_id+5].MaterialType = video::EMT_SOLID;
 }
 
-// To be called once at game init to setup default values.
-void Sky::setSkyDefaults()
+float getWickedTimeOfDay(float time_of_day)
 {
-	SkyboxDefaults sky_defaults;
-	m_sky_params.sky_color = sky_defaults.getSkyColorDefaults();
-	m_sun_params = sky_defaults.getSunDefaults();
-	m_moon_params = sky_defaults.getMoonDefaults();
-	m_star_params = sky_defaults.getStarDefaults();
+	float nightlength = 0.415f;
+	float wn = nightlength / 2;
+	float wicked_time_of_day = 0;
+	if (time_of_day > wn && time_of_day < 1.0f - wn)
+		wicked_time_of_day = (time_of_day - wn) / (1.0f - wn * 2) * 0.5f + 0.25f;
+	else if (time_of_day < 0.5f)
+		wicked_time_of_day = time_of_day / wn * 0.25f;
+	else
+		wicked_time_of_day = 1.0f - ((1.0f - time_of_day) / wn * 0.25f);
+	return wicked_time_of_day;
 }
