@@ -21,24 +21,47 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 #include "util/string.h"
 #include <iostream>
 #include <cstdio>
+#include <cstdlib>
 #include <cstring>
 #include <cerrno>
 #include <fstream>
+#include <atomic>
 #include "log.h"
 #include "config.h"
 #include "porting.h"
-#ifdef __ANDROID__
-#include "settings.h" // For g_settings
+#ifndef SERVER
+#include "irr_ptr.h"
+#include <IFileArchive.h>
+#include <IFileSystem.h>
+#endif
+#ifdef __linux__
+#include <fcntl.h>
+#include <sys/ioctl.h>
+#ifndef FICLONE
+#define FICLONE _IOW(0x94, 9, int)
+#endif
+#endif
+
+// Error from last OS call as string
+#ifdef _WIN32
+#define LAST_OS_ERROR() porting::ConvertError(GetLastError())
+#else
+#define LAST_OS_ERROR() strerror(errno)
 #endif
 
 namespace fs
 {
 
-#ifdef _WIN32 // WINDOWS
+#ifdef _WIN32
+
+/***********
+ * Windows *
+ ***********/
 
-#define _WIN32_WINNT 0x0501
 #include <windows.h>
 #include <shlwapi.h>
+#include <io.h>
+#include <direct.h>
 
 std::vector<DirListNode> GetDirListing(const std::string &pathstring)
 {
@@ -118,6 +141,12 @@ bool IsDir(const std::string &path)
 			(attr & FILE_ATTRIBUTE_DIRECTORY));
 }
 
+bool IsExecutable(const std::string &path)
+{
+	DWORD type;
+	return GetBinaryType(path.c_str(), &type) != 0;
+}
+
 bool IsDirDelimiter(char c)
 {
 	return c == '/' || c == '\\';
@@ -179,16 +208,69 @@ std::string TempPath()
 		errorstream<<"GetTempPath failed, error = "<<GetLastError()<<std::endl;
 		return "";
 	}
-	std::vector<char> buf(bufsize);
+	std::string buf;
+	buf.resize(bufsize);
 	DWORD len = GetTempPath(bufsize, &buf[0]);
 	if(len == 0 || len > bufsize){
 		errorstream<<"GetTempPath failed, error = "<<GetLastError()<<std::endl;
 		return "";
 	}
-	return std::string(buf.begin(), buf.begin() + len);
+	buf.resize(len);
+	return buf;
 }
 
-#else // POSIX
+std::string CreateTempFile()
+{
+	std::string path = TempPath() + DIR_DELIM "MT_XXXXXX";
+	_mktemp_s(&path[0], path.size() + 1); // modifies path
+	HANDLE file = CreateFile(path.c_str(), GENERIC_WRITE, 0, nullptr,
+		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, nullptr);
+	if (file == INVALID_HANDLE_VALUE)
+		return "";
+	CloseHandle(file);
+	return path;
+}
+
+std::string CreateTempDir()
+{
+	std::string path = TempPath() + DIR_DELIM "MT_XXXXXX";
+	_mktemp_s(&path[0], path.size() + 1); // modifies path
+	// will error if it already exists
+	if (!CreateDirectory(path.c_str(), nullptr))
+		return "";
+	return path;
+}
+
+bool CopyFileContents(const std::string &source, const std::string &target)
+{
+	BOOL ok = CopyFileEx(source.c_str(), target.c_str(), nullptr, nullptr,
+		nullptr, COPY_FILE_ALLOW_DECRYPTED_DESTINATION);
+	if (!ok) {
+		errorstream << "copying " << source << " to " << target
+			<< " failed: " << GetLastError() << std::endl;
+		return false;
+	}
+
+	// docs: "File attributes for the existing file are copied to the new file."
+	// This is not our intention so get rid of unwanted attributes:
+	DWORD attr = GetFileAttributes(target.c_str());
+	if (attr == INVALID_FILE_ATTRIBUTES) {
+		errorstream << target << ": file disappeared after copy" << std::endl;
+		return false;
+	}
+	attr &= ~(FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN);
+	SetFileAttributes(target.c_str(), attr);
+
+	tracestream << "copied " << source << " to " << target
+		<< " using CopyFileEx" << std::endl;
+	return true;
+}
+
+#else
+
+/*********
+ * POSIX *
+ *********/
 
 #include <sys/types.h>
 #include <dirent.h>
@@ -282,6 +364,11 @@ bool IsDir(const std::string &path)
 	return ((statbuf.st_mode & S_IFDIR) == S_IFDIR);
 }
 
+bool IsExecutable(const std::string &path)
+{
+	return access(path.c_str(), X_OK) == 0;
+}
+
 bool IsDirDelimiter(char c)
 {
 	return c == '/';
@@ -295,31 +382,26 @@ bool RecursiveDelete(const std::string &path)
 
 	infostream<<"Removing \""<<path<<"\""<<std::endl;
 
-	//return false;
-
 	pid_t child_pid = fork();
 
 	if(child_pid == 0)
 	{
 		// Child
-		char argv_data[3][10000];
+		const char *argv[4] = {
 #ifdef __ANDROID__
-		strcpy(argv_data[0], "/system/bin/rm");
+			"/system/bin/rm",
 #else
-		strcpy(argv_data[0], "/bin/rm");
+			"/bin/rm",
 #endif
-		strcpy(argv_data[1], "-rf");
-		strncpy(argv_data[2], path.c_str(), sizeof(argv_data[2]) - 1);
-		char *argv[4];
-		argv[0] = argv_data[0];
-		argv[1] = argv_data[1];
-		argv[2] = argv_data[2];
-		argv[3] = NULL;
+			"-rf",
+			path.c_str(),
+			NULL
+		};
 
 		verbosestream<<"Executing '"<<argv[0]<<"' '"<<argv[1]<<"' '"
 				<<argv[2]<<"'"<<std::endl;
 
-		execv(argv[0], argv);
+		execv(argv[0], const_cast<char**>(argv));
 
 		// Execv shouldn't return. Failed.
 		_exit(1);
@@ -331,7 +413,6 @@ bool RecursiveDelete(const std::string &path)
 		pid_t tpid;
 		do{
 			tpid = wait(&child_status);
-			//if(tpid != child_pid) process_terminated(tpid);
 		}while(tpid != child_pid);
 		return (child_status == 0);
 	}
@@ -365,15 +446,134 @@ std::string TempPath()
 		compatible with lua's os.tmpname which under the default
 		configuration hardcodes mkstemp("/tmp/lua_XXXXXX").
 	*/
+
 #ifdef __ANDROID__
-	return g_settings->get("TMPFolder");
+	return porting::path_cache;
 #else
 	return DIR_DELIM "tmp";
 #endif
 }
 
+std::string CreateTempFile()
+{
+	std::string path = TempPath() + DIR_DELIM "MT_XXXXXX";
+	int fd = mkstemp(&path[0]); // modifies path
+	if (fd == -1)
+		return "";
+	close(fd);
+	return path;
+}
+
+std::string CreateTempDir()
+{
+	std::string path = TempPath() + DIR_DELIM "MT_XXXXXX";
+	auto r = mkdtemp(&path[0]); // modifies path
+	if (!r)
+		return "";
+	return path;
+}
+
+namespace {
+	struct FileDeleter {
+		void operator()(FILE *stream) {
+			fclose(stream);
+		}
+	};
+
+	typedef std::unique_ptr<FILE, FileDeleter> FileUniquePtr;
+}
+
+bool CopyFileContents(const std::string &source, const std::string &target)
+{
+	FileUniquePtr sourcefile, targetfile;
+
+#ifdef __linux__
+	// Try to clone using Copy-on-Write (CoW). This is instant but supported
+	// only by some filesystems.
+
+	int srcfd, tgtfd;
+	srcfd = open(source.c_str(), O_RDONLY);
+	if (srcfd == -1) {
+		errorstream << source << ": can't open for reading: "
+			<< strerror(errno) << std::endl;
+		return false;
+	}
+	tgtfd = open(target.c_str(), O_WRONLY | O_CREAT | O_TRUNC, 0644);
+	if (tgtfd == -1) {
+		errorstream << target << ": can't open for writing: "
+			<< strerror(errno) << std::endl;
+		close(srcfd);
+		return false;
+	}
+
+	if (ioctl(tgtfd, FICLONE, srcfd) == 0) {
+		tracestream << "copied " << source << " to " << target
+			<< " using FICLONE" << std::endl;
+		close(srcfd);
+		close(tgtfd);
+		return true;
+	}
+
+	// fallback to normal copy, but no need to reopen the files
+	sourcefile.reset(fdopen(srcfd, "rb"));
+	targetfile.reset(fdopen(tgtfd, "wb"));
+	goto fallback;
+
 #endif
 
+	sourcefile.reset(fopen(source.c_str(), "rb"));
+	targetfile.reset(fopen(target.c_str(), "wb"));
+
+fallback:
+
+	if (!sourcefile) {
+		errorstream << source << ": can't open for reading: "
+			<< strerror(errno) << std::endl;
+		return false;
+	}
+	if (!targetfile) {
+		errorstream << target << ": can't open for writing: "
+			<< strerror(errno) << std::endl;
+		return false;
+	}
+
+	size_t total = 0;
+	bool done = false;
+	char readbuffer[BUFSIZ];
+	while (!done) {
+		size_t readbytes = fread(readbuffer, 1,
+				sizeof(readbuffer), sourcefile.get());
+		total += readbytes;
+		if (ferror(sourcefile.get())) {
+			errorstream << source << ": IO error: "
+				<< strerror(errno) << std::endl;
+			return false;
+		}
+		if (readbytes > 0)
+			fwrite(readbuffer, 1, readbytes, targetfile.get());
+		if (feof(sourcefile.get())) {
+			// flush destination file to catch write errors (e.g. disk full)
+			fflush(targetfile.get());
+			done = true;
+		}
+		if (ferror(targetfile.get())) {
+			errorstream << target << ": IO error: "
+					<< strerror(errno) << std::endl;
+			return false;
+		}
+	}
+	tracestream << "copied " << total << " bytes from "
+		<< source << " to " << target << std::endl;
+
+	return true;
+}
+
+#endif
+
+/****************************
+ * portable implementations *
+ ****************************/
+
 void GetRecursiveDirs(std::vector<std::string> &dirs, const std::string &dir)
 {
 	static const std::set<char> chars_to_ignore = { '_', '.' };
@@ -407,21 +607,6 @@ void GetRecursiveSubPaths(const std::string &path,
 	}
 }
 
-bool DeletePaths(const std::vector<std::string> &paths)
-{
-	bool success = true;
-	// Go backwards to succesfully delete the output of GetRecursiveSubPaths
-	for(int i=paths.size()-1; i>=0; i--){
-		const std::string &path = paths[i];
-		bool did = DeleteSingleFileOrEmptyDirectory(path);
-		if(!did){
-			errorstream<<"Failed to delete "<<path<<std::endl;
-			success = false;
-		}
-	}
-	return success;
-}
-
 bool RecursiveDeleteContent(const std::string &path)
 {
 	infostream<<"Removing content of \""<<path<<"\""<<std::endl;
@@ -457,60 +642,6 @@ bool CreateAllDirs(const std::string &path)
 	return true;
 }
 
-bool CopyFileContents(const std::string &source, const std::string &target)
-{
-	FILE *sourcefile = fopen(source.c_str(), "rb");
-	if(sourcefile == NULL){
-		errorstream<<source<<": can't open for reading: "
-			<<strerror(errno)<<std::endl;
-		return false;
-	}
-
-	FILE *targetfile = fopen(target.c_str(), "wb");
-	if(targetfile == NULL){
-		errorstream<<target<<": can't open for writing: "
-			<<strerror(errno)<<std::endl;
-		fclose(sourcefile);
-		return false;
-	}
-
-	size_t total = 0;
-	bool retval = true;
-	bool done = false;
-	char readbuffer[BUFSIZ];
-	while(!done){
-		size_t readbytes = fread(readbuffer, 1,
-				sizeof(readbuffer), sourcefile);
-		total += readbytes;
-		if(ferror(sourcefile)){
-			errorstream<<source<<": IO error: "
-				<<strerror(errno)<<std::endl;
-			retval = false;
-			done = true;
-		}
-		if(readbytes > 0){
-			fwrite(readbuffer, 1, readbytes, targetfile);
-		}
-		if(feof(sourcefile) || ferror(sourcefile)){
-			// flush destination file to catch write errors
-			// (e.g. disk full)
-			fflush(targetfile);
-			done = true;
-		}
-		if(ferror(targetfile)){
-			errorstream<<target<<": IO error: "
-					<<strerror(errno)<<std::endl;
-			retval = false;
-			done = true;
-		}
-	}
-	infostream<<"copied "<<total<<" bytes from "
-		<<source<<" to "<<target<<std::endl;
-	fclose(sourcefile);
-	fclose(targetfile);
-	return retval;
-}
-
 bool CopyDir(const std::string &source, const std::string &target)
 {
 	if(PathExists(source)){
@@ -540,6 +671,30 @@ bool CopyDir(const std::string &source, const std::string &target)
 	return false;
 }
 
+bool MoveDir(const std::string &source, const std::string &target)
+{
+	infostream << "Moving \"" << source << "\" to \"" << target << "\"" << std::endl;
+
+	// If target exists as empty folder delete, otherwise error
+	if (fs::PathExists(target)) {
+		if (rmdir(target.c_str()) != 0) {
+			errorstream << "MoveDir: target \"" << target
+				<< "\" exists as file or non-empty folder" << std::endl;
+			return false;
+		}
+	}
+
+	// Try renaming first which is instant
+	if (fs::Rename(source, target))
+		return true;
+
+	infostream << "MoveDir: rename not possible, will copy instead" << std::endl;
+	bool retval = fs::CopyDir(source, target);
+	if (retval)
+		retval &= fs::RecursiveDelete(source);
+	return retval;
+}
+
 bool PathStartsWith(const std::string &path, const std::string &prefix)
 {
 	size_t pathsize = path.size();
@@ -573,8 +728,8 @@ bool PathStartsWith(const std::string &path, const std::string &prefix)
 				char pathchar = path[pathpos+len];
 				char prefixchar = prefix[prefixpos+len];
 				if(FILESYS_CASE_INSENSITIVE){
-					pathchar = tolower(pathchar);
-					prefixchar = tolower(prefixchar);
+					pathchar = my_tolower(pathchar);
+					prefixchar = my_tolower(prefixchar);
 				}
 				if(pathchar != prefixchar)
 					return false;
@@ -594,7 +749,7 @@ std::string RemoveLastPathComponent(const std::string &path,
 		std::string *removed, int count)
 {
 	if(removed)
-		*removed = "";
+		removed->clear();
 
 	size_t remaining = path.size();
 
@@ -691,52 +846,83 @@ std::string AbsolutePath(const std::string &path)
 const char *GetFilenameFromPath(const char *path)
 {
 	const char *filename = strrchr(path, DIR_DELIM_CHAR);
+	// Consistent with IsDirDelimiter this function handles '/' too
+	if (DIR_DELIM_CHAR != '/') {
+		const char *tmp = strrchr(path, '/');
+		if (tmp && tmp > filename)
+			filename = tmp;
+	}
 	return filename ? filename + 1 : path;
 }
 
-bool safeWriteToFile(const std::string &path, const std::string &content)
+// Note: this is not safe if two MT processes try this at the same time (FIXME?)
+bool safeWriteToFile(const std::string &path, std::string_view content)
 {
-	std::string tmp_file = path + ".~mt";
+	// Prevent two threads from writing to the same temporary file
+	static std::atomic<u16> g_file_counter;
+	const std::string tmp_file = path + ".~mt" + itos(g_file_counter.fetch_add(1));
 
-	// Write to a tmp file
-	std::ofstream os(tmp_file.c_str(), std::ios::binary);
+	// Write data to a temporary file
+	std::string write_error;
+
+#ifdef _WIN32
+	// We've observed behavior suggesting that the MSVC implementation of std::ofstream::flush doesn't
+	// actually flush, so we use win32 APIs.
+	HANDLE handle = CreateFile(tmp_file.c_str(), GENERIC_WRITE, 0, nullptr,
+		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, nullptr);
+	if (handle == INVALID_HANDLE_VALUE) {
+		errorstream << "Failed to open file: " << LAST_OS_ERROR() << std::endl;
+		return false;
+	}
+	DWORD bytes_written;
+	if (!WriteFile(handle, content.data(), content.size(), &bytes_written, nullptr))
+		write_error = LAST_OS_ERROR();
+	else if (!FlushFileBuffers(handle))
+		write_error = LAST_OS_ERROR();
+	CloseHandle(handle);
+#else
+	auto os = open_ofstream(tmp_file.c_str(), true);
 	if (!os.good())
 		return false;
-	os << content;
-	os.flush();
+	os << content << std::flush;
 	os.close();
-	if (os.fail()) {
-		// Remove the temporary file because writing it failed and it's useless.
+	if (os.fail())
+		write_error = "iostream fail";
+#endif
+
+	if (!write_error.empty()) {
+		errorstream << "Failed to write file: " << write_error << std::endl;
 		remove(tmp_file.c_str());
 		return false;
 	}
 
-	bool rename_success = false;
+	std::string rename_error;
 
 	// Move the finished temporary file over the real file
 #ifdef _WIN32
 	// When creating the file, it can cause Windows Search indexer, virus scanners and other apps
 	// to query the file. This can make the move file call below fail.
 	// We retry up to 5 times, with a 1ms sleep between, before we consider the whole operation failed
-	int number_attempts = 0;
-	while (number_attempts < 5) {
-		rename_success = MoveFileEx(tmp_file.c_str(), path.c_str(),
+	for (int attempt = 0; attempt < 5; attempt++) {
+		auto ok = MoveFileEx(tmp_file.c_str(), path.c_str(),
 				MOVEFILE_REPLACE_EXISTING | MOVEFILE_WRITE_THROUGH);
-		if (rename_success)
+		if (ok) {
+			rename_error.clear();
 			break;
+		}
+		rename_error = LAST_OS_ERROR();
 		sleep_ms(1);
-		++number_attempts;
 	}
 #else
 	// On POSIX compliant systems rename() is specified to be able to swap the
 	// file in place of the destination file, making this a truly error-proof
 	// transaction.
-	rename_success = rename(tmp_file.c_str(), path.c_str()) == 0;
+	if (rename(tmp_file.c_str(), path.c_str()) != 0)
+		rename_error = LAST_OS_ERROR();
 #endif
-	if (!rename_success) {
-		warningstream << "Failed to write to file: " << path.c_str() << std::endl;
-		// Remove the temporary file because moving it over the target file
-		// failed.
+
+	if (!rename_error.empty()) {
+		errorstream << "Failed to overwrite \"" << path << "\": " << rename_error << std::endl;
 		remove(tmp_file.c_str());
 		return false;
 	}
@@ -744,10 +930,113 @@ bool safeWriteToFile(const std::string &path, const std::string &content)
 	return true;
 }
 
+#ifndef SERVER
+bool extractZipFile(io::IFileSystem *fs, const char *filename, const std::string &destination)
+{
+	// Be careful here not to touch the global file hierarchy in Irrlicht
+	// since this function needs to be thread-safe!
+
+	io::IArchiveLoader *zip_loader = nullptr;
+	for (u32 i = 0; i < fs->getArchiveLoaderCount(); i++) {
+		if (fs->getArchiveLoader(i)->isALoadableFileFormat(io::EFAT_ZIP)) {
+			zip_loader = fs->getArchiveLoader(i);
+			break;
+		}
+	}
+	if (!zip_loader) {
+		warningstream << "fs::extractZipFile(): Irrlicht said it doesn't support ZIPs." << std::endl;
+		return false;
+	}
+
+	irr_ptr<io::IFileArchive> opened_zip(zip_loader->createArchive(filename, false, false));
+	if (!opened_zip)
+		return false;
+	const io::IFileList* files_in_zip = opened_zip->getFileList();
+
+	for (u32 i = 0; i < files_in_zip->getFileCount(); i++) {
+		if (files_in_zip->isDirectory(i))
+			continue; // ignore, we create dirs as necessary
+
+		const auto &filename = files_in_zip->getFullFileName(i);
+		std::string fullpath = destination + DIR_DELIM;
+		fullpath += filename.c_str();
+
+		fullpath = fs::RemoveRelativePathComponents(fullpath);
+		if (!fs::PathStartsWith(fullpath, destination)) {
+			warningstream << "fs::extractZipFile(): refusing to extract file \""
+				<< filename.c_str() << "\"" << std::endl;
+			continue;
+		}
+
+		std::string fullpath_dir = fs::RemoveLastPathComponent(fullpath);
+
+		if (!fs::PathExists(fullpath_dir) && !fs::CreateAllDirs(fullpath_dir))
+			return false;
+
+		irr_ptr<io::IReadFile> toread(opened_zip->createAndOpenFile(i));
+
+		auto os = open_ofstream(fullpath.c_str(), true);
+		if (!os.good())
+			return false;
+
+		char buffer[4096];
+		long total_read = 0;
+
+		while (total_read < toread->getSize()) {
+			long bytes_read = toread->read(buffer, sizeof(buffer));
+			bool error = true;
+			if (bytes_read != 0) {
+				os.write(buffer, bytes_read);
+				error = os.fail();
+			}
+			if (error) {
+				os.close();
+				remove(fullpath.c_str());
+				return false;
+			}
+			total_read += bytes_read;
+		}
+	}
+
+	return true;
+}
+#endif
+
+bool ReadFile(const std::string &path, std::string &out, bool log_error)
+{
+	auto is = open_ifstream(path.c_str(), log_error, std::ios::ate);
+	if (!is.good())
+		return false;
+
+	auto size = is.tellg();
+	out.resize(size);
+	is.seekg(0);
+	is.read(&out[0], size);
+
+	return !is.fail();
+}
+
 bool Rename(const std::string &from, const std::string &to)
 {
 	return rename(from.c_str(), to.c_str()) == 0;
 }
 
-} // namespace fs
+bool OpenStream(std::filebuf &stream, const char *filename,
+	std::ios::openmode mode, bool log_error, bool log_warn)
+{
+	assert((mode & std::ios::in) || (mode & std::ios::out));
+	assert(!stream.is_open());
+	// C++ dropped the ball hard for file opening error handling, there's not even
+	// an implementation-defined mechanism for returning any kind of error code or message.
+	if (!stream.open(filename, mode)) {
+		if (log_warn || log_error) {
+			std::string msg = LAST_OS_ERROR();
+			(log_error ? errorstream : warningstream)
+				<< "Failed to open \"" << filename << "\": " << msg << std::endl;
+		}
+		return false;
+	}
+	return true;
+}
 
+} // namespace fs
