@@ -22,12 +22,13 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 #include "log.h"
 #include "serverenvironment.h"
 #include "scripting_server.h"
-#include "serverobject.h"
+#include "server/serveractiveobject.h"
 #include "settings.h"
 #include "craftdef.h"
 #include "rollback_interface.h"
 #include "util/strfnd.h"
 #include "util/basic_macros.h"
+#include "inventory.h"
 
 #define PLAYER_TO_SA(p)   p->getEnv()->getScriptIface()
 
@@ -154,44 +155,138 @@ IMoveAction::IMoveAction(std::istream &is, bool somewhere) :
 	}
 }
 
+void IMoveAction::swapDirections()
+{
+	std::swap(from_inv, to_inv);
+	std::swap(from_list, to_list);
+	std::swap(from_i, to_i);
+}
+
+void IMoveAction::onTake(const ItemStack &src_item, ServerActiveObject *player) const
+{
+	ServerScripting *sa = PLAYER_TO_SA(player);
+	if (from_inv.type == InventoryLocation::DETACHED)
+		sa->detached_inventory_OnTake(*this, src_item, player);
+	else if (from_inv.type == InventoryLocation::NODEMETA)
+		sa->nodemeta_inventory_OnTake(*this, src_item, player);
+	else if (from_inv.type == InventoryLocation::PLAYER)
+		sa->player_inventory_OnTake(*this, src_item, player);
+	else
+		assert(false);
+}
+
+void IMoveAction::onPut(const ItemStack &src_item, ServerActiveObject *player) const
+{
+	ServerScripting *sa = PLAYER_TO_SA(player);
+	if (to_inv.type == InventoryLocation::DETACHED)
+		sa->detached_inventory_OnPut(*this, src_item, player);
+	else if (to_inv.type == InventoryLocation::NODEMETA)
+		sa->nodemeta_inventory_OnPut(*this, src_item, player);
+	else if (to_inv.type == InventoryLocation::PLAYER)
+		sa->player_inventory_OnPut(*this, src_item, player);
+	else
+		assert(false);
+}
+
+void IMoveAction::onMove(int count, ServerActiveObject *player) const
+{
+	ServerScripting *sa = PLAYER_TO_SA(player);
+	if (from_inv.type == InventoryLocation::DETACHED)
+		sa->detached_inventory_OnMove(*this, count, player);
+	else if (from_inv.type == InventoryLocation::NODEMETA)
+		sa->nodemeta_inventory_OnMove(*this, count, player);
+	else if (from_inv.type == InventoryLocation::PLAYER)
+		sa->player_inventory_OnMove(*this, count, player);
+	else
+		assert(false);
+}
+
+int IMoveAction::allowPut(const ItemStack &dst_item, ServerActiveObject *player) const
+{
+	ServerScripting *sa = PLAYER_TO_SA(player);
+	int dst_can_put_count = 0xffff;
+	if (to_inv.type == InventoryLocation::DETACHED)
+		dst_can_put_count = sa->detached_inventory_AllowPut(*this, dst_item, player);
+	else if (to_inv.type == InventoryLocation::NODEMETA)
+		dst_can_put_count = sa->nodemeta_inventory_AllowPut(*this, dst_item, player);
+	else if (to_inv.type == InventoryLocation::PLAYER)
+		dst_can_put_count = sa->player_inventory_AllowPut(*this, dst_item, player);
+	else
+		assert(false);
+	return dst_can_put_count;
+}
+
+int IMoveAction::allowTake(const ItemStack &src_item, ServerActiveObject *player) const
+{
+	ServerScripting *sa = PLAYER_TO_SA(player);
+	int src_can_take_count = 0xffff;
+	if (from_inv.type == InventoryLocation::DETACHED)
+		src_can_take_count = sa->detached_inventory_AllowTake(*this, src_item, player);
+	else if (from_inv.type == InventoryLocation::NODEMETA)
+		src_can_take_count = sa->nodemeta_inventory_AllowTake(*this, src_item, player);
+	else if (from_inv.type == InventoryLocation::PLAYER)
+		src_can_take_count = sa->player_inventory_AllowTake(*this, src_item, player);
+	else
+		assert(false);
+	return src_can_take_count;
+}
+
+int IMoveAction::allowMove(int try_take_count, ServerActiveObject *player) const
+{
+	ServerScripting *sa = PLAYER_TO_SA(player);
+	int src_can_take_count = 0xffff;
+	if (from_inv.type == InventoryLocation::DETACHED)
+		src_can_take_count = sa->detached_inventory_AllowMove(*this, try_take_count, player);
+	else if (from_inv.type == InventoryLocation::NODEMETA)
+		src_can_take_count = sa->nodemeta_inventory_AllowMove(*this, try_take_count, player);
+	else if (from_inv.type == InventoryLocation::PLAYER)
+		src_can_take_count = sa->player_inventory_AllowMove(*this, try_take_count, player);
+	else
+		assert(false);
+	return src_can_take_count;
+}
+
 void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGameDef *gamedef)
 {
-	Inventory *inv_from = mgr->getInventory(from_inv);
-	Inventory *inv_to = mgr->getInventory(to_inv);
+	/// Necessary for executing Lua callbacks which may manipulate the inventory,
+	/// hence invalidate pointers needed by IMoveAction::apply
+	auto get_borrow_checked_invlist = [mgr](const InventoryLocation &invloc,
+			const std::string &listname) -> InventoryList::ResizeLocked
+	{
+		Inventory *inv = mgr->getInventory(invloc);
+		if (!inv)
+			return nullptr;
+		InventoryList *list = inv->getList(listname);
+		if (!list)
+			return nullptr;
+		return list->resizeLock();
+	};
 
-	if (!inv_from) {
-		infostream << "IMoveAction::apply(): FAIL: source inventory not found: "
-			<< "from_inv=\""<<from_inv.dump() << "\""
-			<< ", to_inv=\"" << to_inv.dump() << "\"" << std::endl;
-		return;
-	}
-	if (!inv_to) {
-		infostream << "IMoveAction::apply(): FAIL: destination inventory not found: "
-			<< "from_inv=\"" << from_inv.dump() << "\""
-			<< ", to_inv=\"" << to_inv.dump() << "\"" << std::endl;
-		return;
-	}
+	auto list_from = get_borrow_checked_invlist(from_inv, from_list);
+	auto list_to = get_borrow_checked_invlist(to_inv, to_list);
 
-	InventoryList *list_from = inv_from->getList(from_list);
-	InventoryList *list_to = inv_to->getList(to_list);
-
-	/*
-		If a list doesn't exist or the source item doesn't exist
-	*/
 	if (!list_from) {
-		infostream << "IMoveAction::apply(): FAIL: source list not found: "
-			<< "from_inv=\"" << from_inv.dump() << "\""
-			<< ", from_list=\"" << from_list << "\"" << std::endl;
+		infostream << "IMoveAction::apply(): FAIL: source inventory or list not found: "
+				<< "from_inv=\"" << from_inv.dump() << "\""
+				<< ", from_list=\"" << from_list << "\""
+				<< ", to_inv=\"" << to_inv.dump() << "\""
+				<< ", to_list=\"" << to_list << "\""
+				<< std::endl;
 		return;
 	}
 	if (!list_to) {
-		infostream << "IMoveAction::apply(): FAIL: destination list not found: "
-			<< "to_inv=\""<<to_inv.dump() << "\""
-			<< ", to_list=\"" << to_list << "\"" << std::endl;
+		infostream << "IMoveAction::apply(): FAIL: destination inventory or list not found: "
+				<< "from_inv=\"" << from_inv.dump() << "\""
+				<< ", from_list=\"" << from_list << "\""
+				<< ", to_inv=\"" << to_inv.dump() << "\""
+				<< ", to_list=\"" << to_list << "\""
+				<< std::endl;
 		return;
 	}
 
 	if (move_somewhere) {
+		list_from.reset();
+
 		s16 old_to_i = to_i;
 		u16 old_count = count;
 		caused_by_move_somewhere = true;
@@ -209,22 +304,32 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 
 		// Try to add the item to destination list
 		s16 dest_size = list_to->getSize();
+		auto add_to = [&](s16 dest_i) {
+			// release resize lock while the callbacks are happening
+			list_to.reset();
+
+			to_i = dest_i;
+			apply(mgr, player, gamedef);
+			assert(move_count <= count);
+			count -= move_count;
+
+			list_to = get_borrow_checked_invlist(to_inv, to_list);
+			if (!list_to) {
+				// list_to was removed. simulate an empty list
+				dest_size = 0;
+				return;
+			}
+			dest_size = list_to->getSize();
+		};
 		// First try all the non-empty slots
 		for (s16 dest_i = 0; dest_i < dest_size && count > 0; dest_i++) {
-			if (!list_to->getItem(dest_i).empty()) {
-				to_i = dest_i;
-				apply(mgr, player, gamedef);
-				count -= move_count;
-			}
+			if (!list_to->getItem(dest_i).empty())
+				add_to(dest_i);
 		}
-
 		// Then try all the empty ones
 		for (s16 dest_i = 0; dest_i < dest_size && count > 0; dest_i++) {
-			if (list_to->getItem(dest_i).empty()) {
-				to_i = dest_i;
-				apply(mgr, player, gamedef);
-				count -= move_count;
-			}
+			if (list_to->getItem(dest_i).empty())
+				add_to(dest_i);
 		}
 
 		to_i = old_to_i;
@@ -234,12 +339,23 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 		return;
 	}
 
-	if ((u16)to_i > list_to->getSize()) {
-		infostream << "IMoveAction::apply(): FAIL: destination index out of bounds: "
-			<< "to_i=" << to_i
-			<< ", size=" << list_to->getSize() << std::endl;
+	if (from_i < 0 || list_from->getSize() <= (u32) from_i) {
+		infostream << "IMoveAction::apply(): FAIL: source index out of bounds: "
+			<< "size of from_list=\"" << list_from->getSize() << "\""
+			<< ", from_index=\"" << from_i << "\"" << std::endl;
 		return;
 	}
+
+	if (to_i < 0 || list_to->getSize() <= (u32) to_i) {
+		infostream << "IMoveAction::apply(): FAIL: destination index out of bounds: "
+			<< "size of to_list=\"" << list_to->getSize() << "\""
+			<< ", to_index=\"" << to_i << "\"" << std::endl;
+		return;
+	}
+
+	if (list_from.get() == list_to.get() && from_i == to_i)
+		return; // Same slot
+
 	/*
 		Do not handle rollback if both inventories are that of the same player
 	*/
@@ -251,103 +367,95 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 		Collect information of endpoints
 	*/
 
-	int try_take_count = count;
-	if (try_take_count == 0)
-		try_take_count = list_from->getItem(from_i).count;
+	ItemStack src_item = list_from->getItem(from_i);
+	if (count > 0 && count < src_item.count)
+		src_item.count = count;
+	if (src_item.empty())
+		return;
 
 	int src_can_take_count = 0xffff;
 	int dst_can_put_count = 0xffff;
 
-	/* Query detached inventories */
+	// this is needed for swapping items inside one inventory to work
+	ItemStack restitem;
+	bool allow_swap = !list_to->itemFits(to_i, src_item, &restitem)
+		&& restitem.count == src_item.count
+		&& !caused_by_move_somewhere;
+	// move_count : How many items that were moved at the end
+	// count      : Total items "in the queue" of being moved. Do not touch.
+	move_count = src_item.count - restitem.count;
 
-	// Move occurs in the same detached inventory
-	if (from_inv.type == InventoryLocation::DETACHED &&
-			from_inv == to_inv) {
-		src_can_take_count = PLAYER_TO_SA(player)->detached_inventory_AllowMove(
-			*this, try_take_count, player);
-		dst_can_put_count = src_can_take_count;
-	} else {
-		// Destination is detached
-		if (to_inv.type == InventoryLocation::DETACHED) {
-			ItemStack src_item = list_from->getItem(from_i);
-			src_item.count = try_take_count;
-			dst_can_put_count = PLAYER_TO_SA(player)->detached_inventory_AllowPut(
-				*this, src_item, player);
-		}
-		// Source is detached
-		if (from_inv.type == InventoryLocation::DETACHED) {
-			ItemStack src_item = list_from->getItem(from_i);
-			src_item.count = try_take_count;
-			src_can_take_count = PLAYER_TO_SA(player)->detached_inventory_AllowTake(
-				*this, src_item, player);
-		}
+	// Shift-click: Cannot fill this stack, proceed with next slot
+	if (caused_by_move_somewhere && move_count == 0) {
+		return;
 	}
 
-	/* Query node metadata inventories */
+	if (allow_swap) {
+		// Swap will affect the entire stack (= count) if it can performed.
+		src_item = list_from->getItem(from_i);
+		move_count = src_item.count;
+	}
+	src_item.count = move_count; // Temporary movement stack
 
-	// Both endpoints are nodemeta
-	// Move occurs in the same nodemeta inventory
-	if (from_inv.type == InventoryLocation::NODEMETA &&
-			from_inv == to_inv) {
-		src_can_take_count = PLAYER_TO_SA(player)->nodemeta_inventory_AllowMove(
-			*this, try_take_count, player);
-		dst_can_put_count = src_can_take_count;
+	if (from_inv == to_inv) {
+		// Move action within the same inventory
+		src_can_take_count = allowMove(src_item.count, player);
+
+		bool swap_expected = allow_swap;
+		allow_swap = allow_swap
+			&& (src_can_take_count == -1 || src_can_take_count >= src_item.count);
+		if (allow_swap) {
+			int try_put_count = list_to->getItem(to_i).count;
+			swapDirections();
+			dst_can_put_count = allowMove(try_put_count, player);
+			allow_swap = allow_swap
+				&& (dst_can_put_count == -1 || dst_can_put_count >= try_put_count);
+			swapDirections();
+		} else {
+			dst_can_put_count = src_can_take_count;
+		}
+		if (swap_expected != allow_swap)
+			src_can_take_count = dst_can_put_count = 0;
 	} else {
-		// Destination is nodemeta
-		if (to_inv.type == InventoryLocation::NODEMETA) {
-			ItemStack src_item = list_from->getItem(from_i);
-			src_item.count = try_take_count;
-			dst_can_put_count = PLAYER_TO_SA(player)->nodemeta_inventory_AllowPut(
-				*this, src_item, player);
-		}
-		// Source is nodemeta
-		if (from_inv.type == InventoryLocation::NODEMETA) {
-			ItemStack src_item = list_from->getItem(from_i);
-			src_item.count = try_take_count;
-			src_can_take_count = PLAYER_TO_SA(player)->nodemeta_inventory_AllowTake(
-				*this, src_item, player);
+		// Take from one inventory, put into another
+		dst_can_put_count = allowPut(src_item, player);
+		src_can_take_count = allowTake(src_item, player);
+
+		bool swap_expected = allow_swap;
+		allow_swap = allow_swap
+			&& (src_can_take_count == -1 || src_can_take_count >= src_item.count)
+			&& (dst_can_put_count == -1 || dst_can_put_count >= src_item.count);
+		// A swap is expected, which means that we have to
+		// run the "allow" callbacks a second time with swapped inventories
+		if (allow_swap) {
+			ItemStack dst_item = list_to->getItem(to_i);
+			swapDirections();
+
+			int src_can_take = allowPut(dst_item, player);
+			int dst_can_put = allowTake(dst_item, player);
+			allow_swap = allow_swap
+				&& (src_can_take == -1 || src_can_take >= dst_item.count)
+				&& (dst_can_put == -1 || dst_can_put >= dst_item.count);
+			swapDirections();
 		}
+		if (swap_expected != allow_swap)
+			src_can_take_count = dst_can_put_count = 0;
 	}
 
-	// Query player inventories
+	int old_move_count = move_count;
 
-	// Move occurs in the same player inventory
-	if (from_inv.type == InventoryLocation::PLAYER &&
-			from_inv == to_inv) {
-		src_can_take_count = PLAYER_TO_SA(player)->player_inventory_AllowMove(
-			*this, try_take_count, player);
-		dst_can_put_count = src_can_take_count;
-	} else {
-		// Destination is a player
-		if (to_inv.type == InventoryLocation::PLAYER) {
-			ItemStack src_item = list_from->getItem(from_i);
-			src_item.count = try_take_count;
-			dst_can_put_count = PLAYER_TO_SA(player)->player_inventory_AllowPut(
-				*this, src_item, player);
-		}
-		// Source is a player
-		if (from_inv.type == InventoryLocation::PLAYER) {
-			ItemStack src_item = list_from->getItem(from_i);
-			src_item.count = try_take_count;
-			src_can_take_count = PLAYER_TO_SA(player)->player_inventory_AllowTake(
-				*this, src_item, player);
-		}
-	}
+	// Apply limits given by allow_* callbacks
+	if (src_can_take_count != -1)
+		move_count = (u32)std::min<s32>(src_can_take_count, move_count);
+	if (dst_can_put_count != -1)
+		move_count = (u32)std::min<s32>(dst_can_put_count, move_count);
 
-	int old_count = count;
-
-	/* Modify count according to collected data */
-	count = try_take_count;
-	if (src_can_take_count != -1 && count > src_can_take_count)
-		count = src_can_take_count;
-	if (dst_can_put_count != -1 && count > dst_can_put_count)
-		count = dst_can_put_count;
-	/* Limit according to source item count */
-	if (count > list_from->getItem(from_i).count)
-		count = list_from->getItem(from_i).count;
+	// allow_* callbacks should not modify the stack - but if they do - handle that.
+	if (move_count > list_from->getItem(from_i).count)
+		move_count = list_from->getItem(from_i).count;
 
 	/* If no items will be moved, don't go further */
-	if (count == 0) {
+	if (move_count == 0) {
 		// Undo client prediction. See 'clientApply'
 		if (from_inv.type == InventoryLocation::PLAYER)
 			list_from->setModified();
@@ -356,7 +464,7 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 			list_to->setModified();
 
 		infostream<<"IMoveAction::apply(): move was completely disallowed:"
-				<<" count="<<old_count
+				<<" move_count="<<old_move_count
 				<<" from inv=\""<<from_inv.dump()<<"\""
 				<<" list=\""<<from_list<<"\""
 				<<" i="<<from_i
@@ -364,11 +472,11 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 				<<" list=\""<<to_list<<"\""
 				<<" i="<<to_i
 				<<std::endl;
+
 		return;
 	}
 
-	ItemStack src_item = list_from->getItem(from_i);
-	src_item.count = count;
+	// Backups stacks for infinite sources
 	ItemStack from_stack_was = list_from->getItem(from_i);
 	ItemStack to_stack_was = list_to->getItem(to_i);
 
@@ -379,10 +487,13 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 		same as source), nothing happens
 	*/
 	bool did_swap = false;
-	move_count = list_from->moveItem(from_i,
-		list_to, to_i, count, !caused_by_move_somewhere, &did_swap);
+	src_item = list_from->moveItem(from_i,
+		list_to.get(), to_i, move_count, allow_swap, &did_swap);
+	move_count = src_item.count;
 
-	// If source is infinite, reset it's stack
+	assert(allow_swap == did_swap);
+
+	// If source is infinite, reset its stack
 	if (src_can_take_count == -1) {
 		// For the caused_by_move_somewhere == true case we didn't force-put the item,
 		// which guarantees there is no leftover, and code below would duplicate the
@@ -402,23 +513,23 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 			}
 		}
 		if (move_count > 0 || did_swap) {
-			list_from->deleteItem(from_i);
-			list_from->addItem(from_i, from_stack_was);
+			list_from->changeItem(from_i, from_stack_was);
 		}
 	}
-	// If destination is infinite, reset it's stack and take count from source
+	// If destination is infinite, reset its stack and take count from source
 	if (dst_can_put_count == -1) {
-		list_to->deleteItem(to_i);
-		list_to->addItem(to_i, to_stack_was);
-		list_from->deleteItem(from_i);
-		list_from->addItem(from_i, from_stack_was);
-		list_from->takeItem(from_i, count);
+		list_to->changeItem(to_i, to_stack_was);
+		if (did_swap) {
+			// Undo swap result: set the expected stack + size
+			list_from->changeItem(from_i, from_stack_was);
+			list_from->takeItem(from_i, move_count);
+		}
 	}
 
 	infostream << "IMoveAction::apply(): moved"
 			<< " msom=" << move_somewhere
 			<< " caused=" << caused_by_move_somewhere
-			<< " count=" << count
+			<< " move_count=" << move_count
 			<< " from inv=\"" << from_inv.dump() << "\""
 			<< " list=\"" << from_list << "\""
 			<< " i=" << from_i
@@ -428,8 +539,7 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 			<< std::endl;
 
 	// If we are inside the move somewhere loop, we don't need to report
-	// anything if nothing happened (perhaps we don't need to report
-	// anything for caused_by_move_somewhere == true, but this way its safer)
+	// anything if nothing happened
 	if (caused_by_move_somewhere && move_count == 0)
 		return;
 
@@ -470,70 +580,56 @@ void IMoveAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 	/*
 		Report move to endpoints
 	*/
+	list_to.reset();
+	list_from.reset();
 
-	/* Detached inventories */
-
-	// Both endpoints are same detached
-	if (from_inv.type == InventoryLocation::DETACHED &&
-			from_inv == to_inv) {
-		PLAYER_TO_SA(player)->detached_inventory_OnMove(
-				*this, count, player);
+	// Source = destination => move
+	if (from_inv == to_inv) {
+		onMove(move_count, player);
+		if (did_swap) {
+			// Already swapped. The other stack is now placed in "from" list
+			list_from = get_borrow_checked_invlist(from_inv, from_list);
+			if (list_from) {
+				src_item = list_from->getItem(from_i);
+				list_from.reset();
+				swapDirections();
+				onMove(src_item.count, player);
+				swapDirections();
+			}
+		}
+		mgr->setInventoryModified(from_inv);
 	} else {
-		// Destination is detached
-		if (to_inv.type == InventoryLocation::DETACHED) {
-			PLAYER_TO_SA(player)->detached_inventory_OnPut(
-				*this, src_item, player);
+		ItemStack swap_item;
+		if (did_swap) {
+			// Already swapped. The other stack is now placed in "from" list
+			list_from = get_borrow_checked_invlist(from_inv, from_list);
+			if (list_from) {
+				swap_item = list_from->getItem(from_i);
+				list_from.reset();
+			}
 		}
-		// Source is detached
-		if (from_inv.type == InventoryLocation::DETACHED) {
-			PLAYER_TO_SA(player)->detached_inventory_OnTake(
-				*this, src_item, player);
-		}
-	}
 
-	/* Node metadata inventories */
-
-	// Both endpoints are same nodemeta
-	if (from_inv.type == InventoryLocation::NODEMETA &&
-			from_inv == to_inv) {
-		PLAYER_TO_SA(player)->nodemeta_inventory_OnMove(
-			*this, count, player);
-	} else {
-		// Destination is nodemeta
-		if (to_inv.type == InventoryLocation::NODEMETA) {
-			PLAYER_TO_SA(player)->nodemeta_inventory_OnPut(
-				*this, src_item, player);
+		// 1. Take the ItemStack (visually: freely detached)
+		onTake(src_item, player);
+		if (!swap_item.empty() && get_borrow_checked_invlist(to_inv, to_list)) {
+			swapDirections();
+			onTake(swap_item, player);
+			swapDirections();
 		}
-		// Source is nodemeta
-		if (from_inv.type == InventoryLocation::NODEMETA) {
-			PLAYER_TO_SA(player)->nodemeta_inventory_OnTake(
-				*this, src_item, player);
-		}
-	}
 
-	// Player inventories
+		// 2. Put the ItemStack
+		if (get_borrow_checked_invlist(to_inv, to_list))
+			onPut(src_item, player);
 
-	// Both endpoints are same player inventory
-	if (from_inv.type == InventoryLocation::PLAYER &&
-			from_inv == to_inv) {
-		PLAYER_TO_SA(player)->player_inventory_OnMove(
-			*this, count, player);
-	} else {
-		// Destination is player inventory
-		if (to_inv.type == InventoryLocation::PLAYER) {
-			PLAYER_TO_SA(player)->player_inventory_OnPut(
-				*this, src_item, player);
+		if (!swap_item.empty() && get_borrow_checked_invlist(to_inv, to_list)) {
+			swapDirections();
+			onPut(swap_item, player);
+			swapDirections();
 		}
-		// Source is player inventory
-		if (from_inv.type == InventoryLocation::PLAYER) {
-			PLAYER_TO_SA(player)->player_inventory_OnTake(
-				*this, src_item, player);
-		}
-	}
 
-	mgr->setInventoryModified(from_inv);
-	if (inv_from != inv_to)
 		mgr->setInventoryModified(to_inv);
+		mgr->setInventoryModified(from_inv);
+	}
 }
 
 void IMoveAction::clientApply(InventoryManager *mgr, IGameDef *gamedef)
@@ -615,6 +711,8 @@ void IDropAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 		return;
 	}
 
+	auto list_from_lock = list_from->resizeLock();
+
 	/*
 		Do not handle rollback if inventory is player's
 	*/
@@ -695,6 +793,7 @@ void IDropAction::apply(InventoryManager *mgr, ServerActiveObject *player, IGame
 	/*
 		Report drop to endpoints
 	*/
+	list_from_lock.reset();
 
 	switch (from_inv.type) {
 	case InventoryLocation::DETACHED:
@@ -811,6 +910,10 @@ void ICraftAction::apply(InventoryManager *mgr,
 		return;
 	}
 
+	auto list_craft_lock       = list_craft->resizeLock();
+	auto list_craftresult_lock = list_craftresult->resizeLock();
+	auto list_main_lock        = list_main->resizeLock();
+
 	ItemStack crafted;
 	ItemStack craftresultitem;
 	int count_remaining = count;
@@ -871,7 +974,7 @@ void ICraftAction::apply(InventoryManager *mgr,
 		do {
 			PLAYER_TO_SA(player)->item_OnDrop(output_replacement, player,
 				player->getBasePosition());
-			if (count >= output_replacement.count) {
+			if (count <= output_replacement.count) {
 				errorstream << "Couldn't drop replacement stack " <<
 					output_replacement.getItemString() << " because drop loop didn't "
 					"decrease count." << std::endl;
@@ -905,7 +1008,7 @@ bool getCraftingResult(Inventory *inv, ItemStack &result,
 	if (!clist)
 		return false;
 
-	// Mangle crafting grid to an another format
+	// Mangle crafting grid to another format
 	CraftInput ci;
 	ci.method = CRAFT_METHOD_NORMAL;
 	ci.width = clist->getWidth() ? clist->getWidth() : 3;
