@@ -25,7 +25,7 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 
 #include "porting.h"
 
-#if defined(__FreeBSD__)  || defined(__NetBSD__) || defined(__DragonFly__)
+#if defined(__FreeBSD__)  || defined(__NetBSD__) || defined(__DragonFly__) || defined(__OpenBSD__)
 	#include <sys/types.h>
 	#include <sys/sysctl.h>
 	extern char **environ;
@@ -34,25 +34,59 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 	#include <wincrypt.h>
 	#include <algorithm>
 	#include <shlwapi.h>
+	#include <shellapi.h>
+	#include <mmsystem.h>
 #endif
 #if !defined(_WIN32)
 	#include <unistd.h>
 	#include <sys/utsname.h>
+	#if !defined(__ANDROID__)
+		#include <spawn.h>
+	#endif
 #endif
 #if defined(__hpux)
 	#define _PSTAT64
 	#include <sys/pstat.h>
 #endif
+#if defined(__ANDROID__)
+	#include "porting_android.h"
+	#include <android/api-level.h>
+#endif
+#if defined(__APPLE__)
+	#include <mach-o/dyld.h>
+	#include <CoreFoundation/CoreFoundation.h>
+	// For _NSGetEnviron()
+	// Related: https://gitlab.haskell.org/ghc/ghc/issues/2458
+	#include <crt_externs.h>
+#endif
+
+#if defined(__HAIKU__)
+	#include <FindDirectory.h>
+#endif
+
+#if HAVE_MALLOC_TRIM
+	// glibc-only pretty much
+	#include <malloc.h>
+#endif
 
-#include "config.h"
 #include "debug.h"
 #include "filesys.h"
 #include "log.h"
 #include "util/string.h"
-#include "settings.h"
-#include <list>
+#include <vector>
 #include <cstdarg>
 #include <cstdio>
+#include <signal.h>
+#include <atomic>
+
+#if !defined(SERVER) && defined(_WIN32)
+// On Windows export some driver-specific variables to encourage Minetest to be
+// executed on the discrete GPU in case of systems with two. Portability is fun.
+extern "C" {
+	__declspec(dllexport) DWORD NvOptimusEnablement = 1;
+	__declspec(dllexport) DWORD AmdPowerXpressRequestHighPerformance = 1;
+}
+#endif
 
 namespace porting
 {
@@ -61,7 +95,7 @@ namespace porting
 	Signal handler (grabs Ctrl-C on POSIX systems)
 */
 
-bool g_killed = false;
+static bool g_killed = false;
 
 bool *signal_handler_killstatus()
 {
@@ -69,9 +103,8 @@ bool *signal_handler_killstatus()
 }
 
 #if !defined(_WIN32) // POSIX
-	#include <signal.h>
 
-void signal_handler(int sig)
+static void signal_handler(int sig)
 {
 	if (!g_killed) {
 		if (sig == SIGINT) {
@@ -100,9 +133,8 @@ void signal_handler_init(void)
 }
 
 #else // _WIN32
-	#include <signal.h>
 
-BOOL WINAPI event_handler(DWORD sig)
+static BOOL WINAPI event_handler(DWORD sig)
 {
 	switch (sig) {
 	case CTRL_C_EVENT:
@@ -137,11 +169,11 @@ void signal_handler_init(void)
 	Path mangler
 */
 
-// Default to RUN_IN_PLACE style relative paths
-std::string path_share = "..";
-std::string path_user = "..";
-std::string path_locale = path_share + DIR_DELIM + "locale";
-std::string path_cache = path_user + DIR_DELIM + "cache";
+// Nobody should be reading these before initializePaths() is called
+std::string path_share = "UNINITIALIZED";
+std::string path_user = "UNINITIALIZED";
+std::string path_locale = "UNINITIALIZED";
+std::string path_cache = "UNINITIALIZED";
 
 
 std::string getDataPath(const char *subpath)
@@ -149,7 +181,7 @@ std::string getDataPath(const char *subpath)
 	return path_share + DIR_DELIM + subpath;
 }
 
-void pathRemoveFile(char *path, char delim)
+[[maybe_unused]] static void pathRemoveFile(char *path, char delim)
 {
 	// Remove filename and path delimiter
 	int i;
@@ -174,10 +206,12 @@ bool detectMSVCBuildDir(const std::string &path)
 	return (!removeStringEnd(path, ends).empty());
 }
 
-std::string get_sysinfo()
+// Note that the system info is sent in every HTTP request, so keep it reasonably
+// privacy-conserving while ideally still being meaningful.
+
+static std::string detectSystemInfo()
 {
 #ifdef _WIN32
-
 	std::ostringstream oss;
 	LPSTR filePath = new char[MAX_PATH];
 	UINT blockSize;
@@ -197,28 +231,58 @@ std::string get_sysinfo()
 		<< LOWORD(fixedFileInfo->dwProductVersionMS) << '.' // Minor
 		<< HIWORD(fixedFileInfo->dwProductVersionLS) << ' '; // Build
 
-	#ifdef _WIN64
-	oss << "x86_64";
-	#else
-	BOOL is64 = FALSE;
-	if (IsWow64Process(GetCurrentProcess(), &is64) && is64)
-		oss << "x86_64"; // 32-bit app on 64-bit OS
-	else
+	SYSTEM_INFO info;
+	GetNativeSystemInfo(&info);
+	switch (info.wProcessorArchitecture) {
+	case PROCESSOR_ARCHITECTURE_AMD64:
+		oss << "x86_64";
+		break;
+	case PROCESSOR_ARCHITECTURE_ARM:
+		oss << "arm";
+		break;
+	case PROCESSOR_ARCHITECTURE_ARM64:
+		oss << "arm64";
+		break;
+	case PROCESSOR_ARCHITECTURE_INTEL:
 		oss << "x86";
-	#endif
+		break;
+	default:
+		oss << "unknown";
+		break;
+	}
 
 	delete[] lpVersionInfo;
 	delete[] filePath;
 
 	return oss.str();
-#else
+#elif defined(__ANDROID__)
+	std::ostringstream oss;
 	struct utsname osinfo;
 	uname(&osinfo);
-	return std::string(osinfo.sysname) + "/"
-		+ osinfo.release + " " + osinfo.machine;
+	int api = android_get_device_api_level();
+
+	oss << "Android/" << api << " " << osinfo.machine;
+	return oss.str();
+#else /* POSIX */
+	struct utsname osinfo;
+	uname(&osinfo);
+
+	std::string_view release(osinfo.release);
+	// cut off anything but the primary version number
+	release = release.substr(0, release.find_first_not_of("0123456789."));
+
+	std::string ret = osinfo.sysname;
+	ret.append("/").append(release).append(" ").append(osinfo.machine);
+	return ret;
 #endif
 }
 
+const std::string &get_sysinfo()
+{
+	static std::string ret = detectSystemInfo();
+	return ret;
+}
+
 
 bool getCurrentWorkingDir(char *buf, size_t len)
 {
@@ -231,9 +295,9 @@ bool getCurrentWorkingDir(char *buf, size_t len)
 }
 
 
-bool getExecPathFromProcfs(char *buf, size_t buflen)
+static bool getExecPathFromProcfs(char *buf, size_t buflen)
 {
-#ifndef _WIN32
+#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__DragonFly__)
 	buflen--;
 
 	ssize_t len;
@@ -310,6 +374,12 @@ bool getCurrentExecPath(char *buf, size_t len)
 	return true;
 }
 
+#elif defined(__HAIKU__)
+
+bool getCurrentExecPath(char *buf, size_t len)
+{
+	return find_path(B_APP_IMAGE_SYMBOL, B_FIND_PATH_IMAGE_PATH, NULL, buf, len) == B_OK;
+}
 
 //// Solaris
 #elif defined(__sun) || defined(sun)
@@ -354,10 +424,7 @@ bool getCurrentExecPath(char *buf, size_t len)
 #endif
 
 
-//// Non-Windows
-#if !defined(_WIN32)
-
-const char *getHomeOrFail()
+[[maybe_unused]] static inline const char *getHomeOrFail()
 {
 	const char *home = getenv("HOME");
 	// In rare cases the HOME environment variable may be unset
@@ -366,8 +433,6 @@ const char *getHomeOrFail()
 	return home;
 }
 
-#endif
-
 
 //// Windows
 #if defined(_WIN32)
@@ -387,19 +452,33 @@ bool setSystemPaths()
 	path_share = exepath + "\\..";
 	if (detectMSVCBuildDir(exepath)) {
 		// The msvc build dir schould normaly not be present if properly installed,
-		// but its usefull for debugging.
+		// but its useful for debugging.
 		path_share += DIR_DELIM "..";
 	}
 
-	// Use "C:\Users\<user>\AppData\Roaming\<PROJECT_NAME VERSION_MAJOR>"
-	DWORD len = GetEnvironmentVariable("APPDATA", buf, sizeof(buf));
-	FATAL_ERROR_IF(len == 0 || len > sizeof(buf), "Failed to get APPDATA");
+	// Use %MINETEST_USER_PATH%
+	DWORD len = GetEnvironmentVariable("MINETEST_USER_PATH", buf, sizeof(buf));
+	FATAL_ERROR_IF(len > sizeof(buf), "Failed to get MINETEST_USER_PATH (too large for buffer)");
+	if (len == 0) {
+		// Use "C:\Users\<user>\AppData\Roaming\<PROJECT_NAME_C>"
+		len = GetEnvironmentVariable("APPDATA", buf, sizeof(buf));
+		FATAL_ERROR_IF(len == 0 || len > sizeof(buf), "Failed to get APPDATA");
+		path_user = std::string(buf) + DIR_DELIM + PROJECT_NAME_C;
+	} else {
+		path_user = std::string(buf);
+	}
 
-	path_user = std::string(buf) + DIR_DELIM + PROJECT_NAME + std::to_string(VERSION_MAJOR) ;
 	return true;
 }
 
 
+//// Android
+
+#elif defined(__ANDROID__)
+
+extern bool setSystemPaths(); // defined in porting_android.cpp
+
+
 //// Linux
 #elif defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__DragonFly__)
 
@@ -408,11 +487,7 @@ bool setSystemPaths()
 	char buf[BUFSIZ];
 
 	if (!getCurrentExecPath(buf, sizeof(buf))) {
-#ifdef __ANDROID__
-		errorstream << "Unable to read bindir "<< std::endl;
-#else
 		FATAL_ERROR("Unable to read bindir");
-#endif
 		return false;
 	}
 
@@ -421,21 +496,16 @@ bool setSystemPaths()
 
 	// Find share directory from these.
 	// It is identified by containing the subdirectory "builtin".
-	std::list<std::string> trylist;
+	std::vector<std::string> trylist;
 	std::string static_sharedir = STATIC_SHAREDIR;
 	if (!static_sharedir.empty() && static_sharedir != ".")
 		trylist.push_back(static_sharedir);
 
 	trylist.push_back(bindir + DIR_DELIM ".." DIR_DELIM "share"
-		DIR_DELIM + lowercase(PROJECT_NAME) + std::to_string(VERSION_MAJOR));
+		DIR_DELIM + PROJECT_NAME);
 	trylist.push_back(bindir + DIR_DELIM "..");
 
-#ifdef __ANDROID__
-	trylist.push_back(path_user);
-#endif
-
-	for (std::list<std::string>::const_iterator
-			i = trylist.begin(); i != trylist.end(); ++i) {
+	for (auto i = trylist.begin(); i != trylist.end(); ++i) {
 		const std::string &trypath = *i;
 		if (!fs::PathExists(trypath) ||
 			!fs::PathExists(trypath + DIR_DELIM + "builtin")) {
@@ -454,10 +524,13 @@ bool setSystemPaths()
 		break;
 	}
 
-#ifndef __ANDROID__
-	path_user = std::string(getHomeOrFail()) + DIR_DELIM "."
-		+ PROJECT_NAME  + std::to_string(VERSION_MAJOR);
-#endif
+	const char *const minetest_user_path = getenv("MINETEST_USER_PATH");
+	if (minetest_user_path && minetest_user_path[0] != '\0') {
+		path_user = std::string(minetest_user_path);
+	} else {
+		path_user = std::string(getHomeOrFail()) + DIR_DELIM "."
+			+ PROJECT_NAME;
+	}
 
 	return true;
 }
@@ -479,9 +552,14 @@ bool setSystemPaths()
 	}
 	CFRelease(resources_url);
 
-	path_user = std::string(getHomeOrFail())
-		+ "/Library/Application Support/"
-		+ PROJECT_NAME + std::to_string(VERSION_MAJOR);
+	const char *const minetest_user_path = getenv("MINETEST_USER_PATH");
+	if (minetest_user_path && minetest_user_path[0] != '\0') {
+		path_user = std::string(minetest_user_path);
+	} else {
+		path_user = std::string(getHomeOrFail())
+			+ "/Library/Application Support/"
+			+ PROJECT_NAME;
+	}
 	return true;
 }
 
@@ -491,15 +569,21 @@ bool setSystemPaths()
 bool setSystemPaths()
 {
 	path_share = STATIC_SHAREDIR;
-	path_user  = std::string(getHomeOrFail()) + DIR_DELIM "."
-		+ lowercase(PROJECT_NAME) + std::to_string(VERSION_MAJOR);
+	const char *const minetest_user_path = getenv("MINETEST_USER_PATH");
+	if (minetest_user_path && minetest_user_path[0] != '\0') {
+		path_user = std::string(minetest_user_path);
+	} else {
+		path_user  = std::string(getHomeOrFail()) + DIR_DELIM "."
+			+ lowercase(PROJECT_NAME);
+	}
 	return true;
 }
 
 
 #endif
 
-void migrateCachePath()
+// Move cache folder from path_user to system cache location if possible.
+[[maybe_unused]] static void migrateCachePath()
 {
 	const std::string local_cache_path = path_user + DIR_DELIM + "cache";
 
@@ -519,13 +603,30 @@ void migrateCachePath()
 	}
 }
 
+// Create tag in cache folder according to <https://bford.info/cachedir/> spec
+static void createCacheDirTag()
+{
+	const auto path = path_cache + DIR_DELIM + "CACHEDIR.TAG";
+
+	if (fs::PathExists(path))
+		return;
+	fs::CreateAllDirs(path_cache);
+	auto ofs = open_ofstream(path.c_str(), false);
+	if (!ofs.good())
+		return;
+	ofs << "Signature: 8a477f597d28d172789f06886806bc55\n"
+		"# This file is a cache directory tag automatically created by "
+		PROJECT_NAME_C ".\n"
+		"# For information about cache directory tags, see: "
+		"https://bford.info/cachedir/\n";
+}
+
 void initializePaths()
 {
 #if RUN_IN_PLACE
-	char buf[BUFSIZ];
-
 	infostream << "Using relative paths (RUN_IN_PLACE)" << std::endl;
 
+	char buf[BUFSIZ];
 	bool success =
 		getCurrentExecPath(buf, sizeof(buf)) ||
 		getExecPathFromProcfs(buf, sizeof(buf));
@@ -563,39 +664,44 @@ void initializePaths()
 		path_user  = execpath;
 	}
 	path_cache = path_user + DIR_DELIM + "cache";
+
 #else
 	infostream << "Using system-wide paths (NOT RUN_IN_PLACE)" << std::endl;
 
 	if (!setSystemPaths())
 		errorstream << "Failed to get one or more system-wide path" << std::endl;
 
-
-#  ifdef _WIN32
+#  ifdef __ANDROID__
+	sanity_check(!path_cache.empty());
+#  elif defined(_WIN32)
 	path_cache = path_user + DIR_DELIM + "cache";
 #  else
-	// Initialize path_cache
-	// First try $XDG_CACHE_HOME/PROJECT_NAME+VERSION_MAJOR
+	// First try $XDG_CACHE_HOME/PROJECT_NAME
 	const char *cache_dir = getenv("XDG_CACHE_HOME");
 	const char *home_dir = getenv("HOME");
-	if (cache_dir) {
-		path_cache = std::string(cache_dir) + DIR_DELIM + lowercase(PROJECT_NAME) + std::to_string(VERSION_MAJOR);
+	if (cache_dir && cache_dir[0] != '\0') {
+		path_cache = std::string(cache_dir) + DIR_DELIM + PROJECT_NAME;
 	} else if (home_dir) {
-		// Then try $HOME/.cache/PROJECT_NAME+VERSION_MAJOR
+		// Then try $HOME/.cache/PROJECT_NAME
 		path_cache = std::string(home_dir) + DIR_DELIM + ".cache"
-			+ DIR_DELIM + lowercase(PROJECT_NAME) + std::to_string(VERSION_MAJOR);
+			+ DIR_DELIM + PROJECT_NAME;
 	} else {
 		// If neither works, use $PATH_USER/cache
 		path_cache = path_user + DIR_DELIM + "cache";
 	}
+#  endif // _WIN32
+
 	// Migrate cache folder to new location if possible
 	migrateCachePath();
-#  endif // _WIN32
+
 #endif // RUN_IN_PLACE
 
 	infostream << "Detected share path: " << path_share << std::endl;
 	infostream << "Detected user path: " << path_user << std::endl;
 	infostream << "Detected cache path: " << path_cache << std::endl;
 
+	createCacheDirTag();
+
 #if USE_GETTEXT
 	bool found_localedir = false;
 #  ifdef STATIC_LOCALEDIR
@@ -663,19 +769,83 @@ bool secure_rand_fill_buf(void *buf, size_t len)
 
 #endif
 
+#ifndef __ANDROID__
+
+void osSpecificInit()
+{
+#ifdef _WIN32
+	// hardening options
+	HeapSetInformation(NULL, HeapEnableTerminationOnCorruption, NULL, 0);
+	SetSearchPathMode(BASE_SEARCH_PATH_ENABLE_SAFE_SEARCHMODE |
+		BASE_SEARCH_PATH_PERMANENT);
+	SetProcessDEPPolicy(PROCESS_DEP_ENABLE);
+#endif
+}
+
+#endif
+
 void attachOrCreateConsole()
 {
 #ifdef _WIN32
-	static bool consoleAllocated = false;
-	const bool redirected = (_fileno(stdout) == -2 || _fileno(stdout) == -1); // If output is redirected to e.g a file
-	if (!consoleAllocated && redirected && (AttachConsole(ATTACH_PARENT_PROCESS) || AllocConsole())) {
-		freopen("CONOUT$", "w", stdout);
-		freopen("CONOUT$", "w", stderr);
-		consoleAllocated = true;
+	static bool once = false;
+	const bool redirected = _fileno(stdout) >= 0; // If output is redirected to e.g a file
+	if (!once && !redirected) {
+		if (AttachConsole(ATTACH_PARENT_PROCESS) || AllocConsole()) {
+			freopen("CONOUT$", "w", stdout);
+			freopen("CONOUT$", "w", stderr);
+		}
+		once = true;
 	}
 #endif
 }
 
+#ifdef _WIN32
+std::string QuoteArgv(const std::string &arg)
+{
+	// Quoting rules on Windows are batshit insane, can differ between applications
+	// and there isn't even a stdlib function to deal with it.
+	// Ref: https://learn.microsoft.com/archive/blogs/twistylittlepassagesallalike/everyone-quotes-command-line-arguments-the-wrong-way
+	if (!arg.empty() && arg.find_first_of(" \t\n\v\"") == std::string::npos)
+		return arg;
+
+	std::string ret;
+	ret.reserve(arg.size()+2);
+	ret.push_back('"');
+	for (auto it = arg.begin(); it != arg.end(); ++it) {
+		u32 back = 0;
+		while (it != arg.end() && *it == '\\')
+			++back, ++it;
+
+		if (it == arg.end()) {
+			ret.append(2 * back, '\\');
+			break;
+		} else if (*it == '"') {
+			ret.append(2 * back + 1, '\\');
+		} else {
+			ret.append(back, '\\');
+		}
+		ret.push_back(*it);
+	}
+	ret.push_back('"');
+	return ret;
+}
+
+std::string ConvertError(DWORD error_code)
+{
+	wchar_t buffer[320];
+
+	auto r = FormatMessageW(
+		FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
+		nullptr, error_code, 0, buffer, ARRLEN(buffer) - 1, nullptr);
+	if (!r)
+		return std::to_string(error_code);
+
+	if (!buffer[0]) // should not happen normally
+		return "?";
+	return wide_to_utf8(buffer);
+}
+#endif
+
 int mt_snprintf(char *buf, const size_t buf_size, const char *fmt, ...)
 {
 	// https://msdn.microsoft.com/en-us/library/bt7tawza.aspx
@@ -698,11 +868,61 @@ int mt_snprintf(char *buf, const size_t buf_size, const char *fmt, ...)
 	return c;
 }
 
+#ifdef __ANDROID__
+// defined in porting_android.cpp
+extern void openURIAndroid(const char *url);
+#endif
+
+static bool open_uri(const std::string &uri)
+{
+	if (uri.find_first_of("\r\n") != std::string::npos) {
+		errorstream << "Unable to open URI as it is invalid, contains new line: " << uri << std::endl;
+		return false;
+	}
+
+#if defined(_WIN32)
+	return (intptr_t)ShellExecuteA(NULL, NULL, uri.c_str(), NULL, NULL, SW_SHOWNORMAL) > 32;
+#elif defined(__ANDROID__)
+	openURIAndroid(uri.c_str());
+	return true;
+#elif defined(__APPLE__)
+	const char *argv[] = {"open", uri.c_str(), NULL};
+	return posix_spawnp(NULL, "open", NULL, NULL, (char**)argv,
+		(*_NSGetEnviron())) == 0;
+#else
+	const char *argv[] = {"xdg-open", uri.c_str(), NULL};
+	return posix_spawnp(NULL, "xdg-open", NULL, NULL, (char**)argv, environ) == 0;
+#endif
+}
+
+bool open_url(const std::string &url)
+{
+	if (!str_starts_with(url, "http://") && !str_starts_with(url, "https://")) {
+		errorstream << "Unable to open browser as URL is missing schema: " << url << std::endl;
+		return false;
+	}
+
+	return open_uri(url);
+}
+
+bool open_directory(const std::string &path)
+{
+	if (!fs::IsDir(path)) {
+		errorstream << "Unable to open directory as it does not exist: " << path << std::endl;
+		return false;
+	}
+
+	return open_uri(path);
+}
+
 // Load performance counter frequency only once at startup
 #ifdef _WIN32
 
 inline double get_perf_freq()
 {
+	// Also use this opportunity to enable high-res timers
+	timeBeginPeriod(1);
+
 	LARGE_INTEGER freq;
 	QueryPerformanceFrequency(&freq);
 	return freq.QuadPart;
@@ -712,4 +932,44 @@ double perf_freq = get_perf_freq();
 
 #endif
 
+#if HAVE_MALLOC_TRIM
+
+/*
+ * On Linux/glibc we found that after deallocating bigger chunks of data (esp. MapBlocks)
+ * the memory would not be given back to the OS and would stay at peak usage.
+ * This appears to be a combination of unfortunate allocation order/fragmentation
+ * and the fact that glibc does not call madvise(MADV_DONTNEED) on its own.
+ * Some other allocators were also affected, jemalloc and musl libc were not.
+ * read more: <https://forum.minetest.net/viewtopic.php?t=30509>
+ *
+ * As a workaround we track freed memory coarsely and call malloc_trim() once a
+ * certain amount is reached.
+ *
+ * Because trimming can take more than 10ms and would cause jitter if done
+ * uncontrolled we have a separate function, which is called from background threads.
+ */
+
+static std::atomic<size_t> memory_freed;
+
+constexpr size_t MEMORY_TRIM_THRESHOLD = 256 * 1024 * 1024;
+
+void TrackFreedMemory(size_t amount)
+{
+	memory_freed.fetch_add(amount, std::memory_order_relaxed);
+}
+
+void TriggerMemoryTrim()
+{
+	constexpr auto MO = std::memory_order_relaxed;
+	if (memory_freed.load(MO) >= MEMORY_TRIM_THRESHOLD) {
+		// Synchronize call
+		if (memory_freed.exchange(0, MO) < MEMORY_TRIM_THRESHOLD)
+			return;
+		// Leave some headroom for future allocations
+		malloc_trim(8 * 1024 * 1024);
+	}
+}
+
+#endif
+
 } //namespace porting
